# Copyright (c) 2019-2021, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ezdxf.eztypes import Drawing, DXFEntity, ExtendedTags

__all__ = [
    "register_entity",
    "ENTITY_CLASSES",
    "replace_entity",
    "new",
    "cls",
    "is_bound",
    "create_db_entry",
    "load",
    "bind",
]
# Stores all registered classes:
ENTITY_CLASSES = {}
# use @set_default_class to register the default entity class:
DEFAULT_CLASS = None


def set_default_class(cls):
    global DEFAULT_CLASS
    DEFAULT_CLASS = cls
    return cls


def replace_entity(cls):
    name = cls.DXFTYPE
    ENTITY_CLASSES[name] = cls
    return cls


def register_entity(cls):
    name = cls.DXFTYPE
    if name in ENTITY_CLASSES:
        raise TypeError(f"Double registration for DXF type {name}.")
    ENTITY_CLASSES[name] = cls
    return cls


def new(
    dxftype: str, dxfattribs: dict = None, doc: "Drawing" = None
) -> "DXFEntity":
    """Create a new entity, does not require an instantiated DXF document."""
    entity = cls(dxftype).new(
        handle=None,
        owner=None,
        dxfattribs=dxfattribs,
        doc=doc,
    )
    return entity.cast() if hasattr(entity, "cast") else entity  # type: ignore


def create_db_entry(dxftype, dxfattribs: dict, doc: "Drawing") -> "DXFEntity":
    entity = new(dxftype=dxftype, dxfattribs=dxfattribs)
    bind(entity, doc)
    return entity


def load(tags: "ExtendedTags", doc: "Drawing" = None) -> "DXFEntity":
    entity = cls(tags.dxftype()).load(tags, doc)
    return entity.cast() if hasattr(entity, "cast") else entity  # type: ignore


def cls(dxftype: str) -> "DXFEntity":
    """Returns registered class for `dxftype`."""
    return ENTITY_CLASSES.get(dxftype, DEFAULT_CLASS)


def bind(entity: "DXFEntity", doc: "Drawing") -> None:
    """Bind `entity` to the DXF document `doc`.

    The bind process stores the DXF `entity` in the entity database of the DXF
    document.

    """
    assert entity.is_alive, "Can not bind destroyed entity."
    assert doc.entitydb is not None, "Missing entity database."
    entity.doc = doc
    doc.entitydb.add(entity)

    # Do not call the post_bind_hook() while loading from external sources,
    # not all entities and resources are loaded at this point of time!
    if not doc.is_loading:
        # bind extension dictionary
        if entity.extension_dict is not None:
            xdict = entity.extension_dict
            if xdict.has_valid_dictionary:
                xdict.update_owner(entity.dxf.handle)
                dictionary = xdict.dictionary
                if not is_bound(dictionary, doc):
                    bind(dictionary, doc)
                    doc.objects.add_object(dictionary)
        entity.post_bind_hook()


def unbind(entity: "DXFEntity"):
    """Unbind `entity` from document and layout, but does not destroy the
    entity.

    Turns `entity` into a virtual entity: no handle, no owner, no document.
    """
    if entity.is_alive and not entity.is_virtual:
        doc = entity.doc
        if entity.dxf.owner is not None:
            try:
                layout = doc.layouts.get_layout_for_entity(entity)  # type: ignore
            except KeyError:
                pass
            else:
                layout.unlink_entity(entity)  # type: ignore

        process_sub_entities = getattr(entity, "process_sub_entities", None)
        if process_sub_entities:
            process_sub_entities(lambda e: unbind(e))

        doc.entitydb.discard(entity)  # type: ignore
        entity.doc = None


def is_bound(entity: "DXFEntity", doc: "Drawing") -> bool:
    """Returns ``True`` if `entity`is bound to DXF document `doc`."""
    if not entity.is_alive:
        return False
    if entity.is_virtual or entity.doc is not doc:
        return False
    assert doc.entitydb, "Missing entity database."
    return entity.dxf.handle in doc.entitydb



# Dummy file
# features of v0.15.1
from ezdxf.path import Path, Command
import warnings

warnings.warn(
    'Import class Path() only from ezdxf.path, ezdxf.render.path.py will be '
    'removed in v0.17.', DeprecationWarning)



# Copyright (C) 2018-2021, Manfred Moitzi
# License: MIT License
import logging
import io
import bisect
import math
from typing import TextIO, Iterable, List, Optional, Set, cast

from .const import (
    DXFStructureError,
    DXFError,
    DXFValueError,
    DXFAppDataError,
    DXFXDataError,
    APP_DATA_MARKER,
    HEADER_VAR_MARKER,
    XDATA_MARKER,
    INVALID_LAYER_NAME_CHARACTERS,
    acad_release,
    VALID_DXF_LINEWEIGHT_VALUES,
    VALID_DXF_LINEWEIGHTS,
    LINEWEIGHT_BYLAYER,
)

from .tagger import ascii_tags_loader
from .types import is_embedded_object_marker, DXFTag, NONE_TAG
from ezdxf.tools.codepage import toencoding
from ezdxf.math import NULLVEC

logger = logging.getLogger("ezdxf")


class DXFInfo:
    def __init__(self):
        self.release = "R12"
        self.version = "AC1009"
        self.encoding = "cp1252"
        self.handseed = "0"

    def set_header_var(self, name: str, value: str) -> int:
        if name == "$ACADVER":
            self.version = value
            self.release = acad_release.get(value, "R12")
        elif name == "$DWGCODEPAGE":
            self.encoding = toencoding(value)
        elif name == "$HANDSEED":
            self.handseed = value
        else:
            return 0
        return 1


def dxf_info(stream: TextIO) -> DXFInfo:
    info = DXFInfo()
    tagger = ascii_tags_loader(stream)
    # comments already removed
    if next(tagger) != (0, "SECTION"):
        # maybe a DXF structure error, handled by later processing
        return info
    if next(tagger) != (2, "HEADER"):
        # no leading HEADER section like DXF R12 with only ENTITIES section
        return info
    tag = NONE_TAG
    found: int = 0
    while tag != (0, "ENDSEC"):  # until end of HEADER section
        tag = next(tagger)
        if tag.code != HEADER_VAR_MARKER:
            continue
        name = cast(str, tag.value)
        value = cast(str, next(tagger).value)
        found += info.set_header_var(name, value)
        if found > 2:  # all expected values collected
            break
    return info


def header_validator(tagger: Iterable[DXFTag]) -> Iterable[DXFTag]:
    """Checks the tag structure of the content of the header section.

    Do not feed (0, 'SECTION') (2, 'HEADER') and (0, 'ENDSEC') tags!

    Args:
        tagger: generator/iterator of low level tags or compiled tags

    Raises:
        DXFStructureError() -> invalid group codes
        DXFValueError() -> invalid header variable name

    """
    variable_name_tag = True
    for tag in tagger:
        code, value = tag
        if variable_name_tag:
            if code != HEADER_VAR_MARKER:
                raise DXFStructureError(
                    f"Invalid header variable tag {code}, {value})."
                )
            if not value.startswith("$"):
                raise DXFValueError(
                    f'Invalid header variable name "{value}", missing leading "$".'
                )
            variable_name_tag = False
        else:
            variable_name_tag = True
        yield tag


def entity_structure_validator(tags: List[DXFTag]) -> Iterable[DXFTag]:
    """Checks for valid DXF entity tag structure.

    - APP DATA can not be nested and every opening tag (102, '{...') needs a
      closing tag (102, '}')
    - extended group codes (>=1000) allowed before XDATA section
    - XDATA section starts with (1001, APPID) and is always at the end of an
      entity
    - XDATA section: only group code >= 1000 is allowed
    - XDATA control strings (1002, '{') and (1002, '}') have to be balanced
    - embedded objects may follow XDATA

    XRECORD entities will not be checked.

    Args:
        tags: list of DXFTag()

    Raises:
        DXFAppDataError: for invalid APP DATA
        DXFXDataError: for invalid XDATA

    """
    assert isinstance(tags, list)
    dxftype = cast(str, tags[0].value)
    is_xrecord = dxftype == "XRECORD"
    handle: str = "???"
    app_data: bool = False
    xdata: bool = False
    xdata_list_level: int = 0
    app_data_closing_tag: str = "}"
    embedded_object: bool = False
    for tag in tags:
        if tag.code == 5 and handle == "???":
            handle = cast(str, tag.value)

        if is_embedded_object_marker(tag):
            embedded_object = True

        if embedded_object:  # no further validation
            yield tag
            continue  # with next tag

        if xdata and not embedded_object:
            if tag.code < 1000:
                dxftype = cast(str, tags[0].value)
                raise DXFXDataError(
                    f"Invalid XDATA structure in entity {dxftype}(#{handle}), "
                    f"only group code >=1000 allowed in XDATA section"
                )
            if tag.code == 1002:
                value = cast(str, tag.value)
                if value == "{":
                    xdata_list_level += 1
                elif value == "}":
                    xdata_list_level -= 1
                else:
                    raise DXFXDataError(
                        f'Invalid XDATA control string (1002, "{value}") entity'
                        f" {dxftype}(#{handle})."
                    )
                if xdata_list_level < 0:  # more closing than opening tags
                    raise DXFXDataError(
                        f"Invalid XDATA structure in entity {dxftype}(#{handle}), "
                        f'unbalanced list markers, missing  (1002, "{{").'
                    )

        if tag.code == APP_DATA_MARKER and not is_xrecord:
            # Ignore control tags (102, ...) tags in XRECORD
            value = cast(str, tag.value)
            if value.startswith("{"):
                if app_data:  # already in app data mode
                    raise DXFAppDataError(
                        f"Invalid APP DATA structure in entity {dxftype}"
                        f"(#{handle}), APP DATA can not be nested."
                    )
                app_data = True
                # 'APPID}' is also a valid closing tag
                app_data_closing_tag = value[1:] + "}"
            elif value == "}" or value == app_data_closing_tag:
                if not app_data:
                    raise DXFAppDataError(
                        f"Invalid APP DATA structure in entity {dxftype}"
                        f'(#{handle}), found (102, "}}") tag without opening tag.'
                    )
                app_data = False
                app_data_closing_tag = "}"
            else:
                raise DXFAppDataError(
                    f'Invalid APP DATA structure tag (102, "{value}") in '
                    f"entity {dxftype}(#{handle})."
                )

        # XDATA section starts with (1001, APPID) and is always at the end of
        # an entity.
        if tag.code == XDATA_MARKER and xdata is False:
            xdata = True
            if app_data:
                raise DXFAppDataError(
                    f"Invalid APP DATA structure in entity {dxftype}"
                    f'(#{handle}), missing closing tag (102, "}}").'
                )
        yield tag

    if app_data:
        raise DXFAppDataError(
            f"Invalid APP DATA structure in entity {dxftype}(#{handle}), "
            f'missing closing tag (102, "}}").'
        )
    if xdata:
        if xdata_list_level < 0:
            raise DXFXDataError(
                f"Invalid XDATA structure in entity {dxftype}(#{handle}), "
                f'unbalanced list markers, missing  (1002, "{{").'
            )
        elif xdata_list_level > 0:
            raise DXFXDataError(
                f"Invalid XDATA structure in entity {dxftype}(#{handle}), "
                f'unbalanced list markers, missing  (1002, "}}").'
            )


def is_dxf_file(filename: str) -> bool:
    """Returns ``True`` if `filename` is an ASCII DXF file."""
    with io.open(filename, errors="ignore") as fp:
        return is_dxf_stream(fp)


def is_binary_dxf_file(filename: str) -> bool:
    """Returns ``True`` if `filename` is a binary DXF file."""
    with open(filename, "rb") as fp:
        sentinel = fp.read(22)
    return sentinel == b"AutoCAD Binary DXF\r\n\x1a\x00"


def is_dwg_file(filename: str) -> bool:
    """Returns ``True`` if `filename` is a DWG file."""
    return dwg_version(filename) is not None


def dwg_version(filename: str) -> Optional[str]:
    """Returns DWG version of `filename` as string or ``None``."""
    with open(str(filename), "rb") as fp:
        try:
            version = fp.read(6).decode(errors="ignore")
        except IOError:
            return None
        if version not in acad_release:
            return None
        return version


def is_dxf_stream(stream: TextIO) -> bool:
    try:
        reader = ascii_tags_loader(stream)
    except DXFError:
        return False
    try:
        for tag in reader:
            # The common case for well formed DXF files
            if tag == (0, "SECTION"):
                return True
            # Accept/Ignore tags in front of first SECTION - like AutoCAD and
            # BricsCAD, but group code should be < 1000, until reality proofs
            # otherwise.
            if tag.code > 999:
                return False
    except DXFStructureError:
        pass
    return False


# Names used in symbol table records and in dictionaries must follow these rules:
#
# - Names can be any length in ObjectARX, but symbol names entered by users in
#   AutoCAD are limited to 255 characters.
# - AutoCAD preserves the case of names but does not use the case in
#   comparisons. For example, AutoCAD considers "Floor" to be the same symbol
#   as "FLOOR."
# - Names can be composed of all characters allowed by Windows or Mac OS for
#   filenames, except comma (,), backquote (‘), semi-colon (;), and equal
#   sign (=).
# http://help.autodesk.com/view/OARX/2018/ENU/?guid=GUID-83ABF20A-57D4-4AB3-8A49-D91E0F70DBFF


def is_valid_table_name(name: str) -> bool:
    return not bool(INVALID_LAYER_NAME_CHARACTERS.intersection(set(name)))


def is_valid_layer_name(name: str) -> bool:
    if is_adsk_special_layer(name):
        return True
    else:
        return is_valid_table_name(name)


def is_adsk_special_layer(name: str) -> bool:
    if name.startswith("*"):
        # special Autodesk layers starts with invalid character *
        name = name.upper()
        if name.startswith("*ADSK_"):
            return True
        if name.startswith("*ACMAP"):
            return True
        if name.startswith("*TEMPORARY"):
            return True
    return False


def is_valid_block_name(name: str) -> bool:
    if name.startswith("*"):
        return is_valid_table_name(name[1:])
    else:
        return is_valid_table_name(name)


def is_valid_vport_name(name: str) -> bool:
    if name.startswith("*"):
        return name.upper() == "*ACTIVE"
    else:
        return is_valid_table_name(name)


def is_valid_lineweight(lineweight: int) -> bool:
    return lineweight in VALID_DXF_LINEWEIGHT_VALUES


def fix_lineweight(lineweight: int) -> int:
    if lineweight in VALID_DXF_LINEWEIGHT_VALUES:
        return lineweight
    if lineweight < -3:
        return LINEWEIGHT_BYLAYER
    if lineweight > 211:
        return 211
    index = bisect.bisect(VALID_DXF_LINEWEIGHTS, lineweight)
    return VALID_DXF_LINEWEIGHTS[index]


def is_valid_aci_color(aci: int) -> bool:
    return 0 <= aci <= 257


def is_in_integer_range(start: int, end: int):
    """Range of integer values, excluding the `end` value."""

    def _validator(value: int) -> bool:
        return start <= value < end

    return _validator


def fit_into_integer_range(start: int, end: int):
    def _fixer(value: int) -> int:
        return min(max(value, start), end - 1)

    return _fixer


def fit_into_float_range(start: float, end: float):
    def _fixer(value: float) -> float:
        return min(max(value, start), end)

    return _fixer


def is_in_float_range(start: float, end: float):
    """Range of float values, including the `end` value."""

    def _validator(value: float) -> bool:
        return start <= value <= end

    return _validator


def is_not_null_vector(v) -> bool:
    return not NULLVEC.isclose(v)


def is_not_zero(v: float) -> bool:
    return not math.isclose(v, 0.0, abs_tol=1e-12)


def is_not_negative(v) -> bool:
    return v >= 0


is_greater_or_equal_zero = is_not_negative


def is_positive(v) -> bool:
    return v > 0


is_greater_zero = is_positive


def is_valid_bitmask(mask: int):
    def _validator(value: int) -> bool:
        return not bool(~mask & value)

    return _validator


def fix_bitmask(mask: int):
    def _fixer(value: int) -> int:
        return mask & value

    return _fixer


def is_integer_bool(v) -> bool:
    return v in (0, 1)


def fix_integer_bool(v) -> int:
    return 1 if v else 0


def is_one_of(values: Set):
    def _validator(v) -> bool:
        return v in values

    return _validator


def is_valid_one_line_text(text: str) -> bool:
    has_line_breaks = bool(set(text).intersection({"\n", "\r"}))
    return not has_line_breaks and not text.endswith("^")


def fix_one_line_text(text: str) -> str:
    return text.replace("\n", "").replace("\r", "").rstrip("^")


def is_valid_attrib_tag(tag: str) -> bool:
    return is_valid_one_line_text(tag)


def fix_attrib_tag(tag: str) -> str:
    return fix_one_line_text(tag)


def is_handle(handle) -> bool:
    try:
        int(handle, 16)
    except (ValueError, TypeError):
        return False
    return True



# Copyright (c) 2010-2021 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Sequence
import math
from ezdxf.math import Vec2
from .line import ConstructionRay
from .bbox import BoundingBox2d

if TYPE_CHECKING:
    from ezdxf.eztypes import Vertex

HALF_PI = math.pi / 2.0

__all__ = ["ConstructionCircle"]


class ConstructionCircle:
    """Circle construction tool.

    Args:
        center: center point as :class:`Vec2` compatible object
        radius: circle radius > `0`

    """

    def __init__(self, center: "Vertex", radius: float = 1.0):
        self.center = Vec2(center)
        self.radius = float(radius)
        if self.radius <= 0.0:
            raise ValueError("Radius has to be > 0.")

    def __str__(self) -> str:
        """Returns string representation of circle
        "ConstructionCircle(center, radius)".
        """
        return f"ConstructionCircle({self.center}, {self.radius})"

    @staticmethod
    def from_3p(
        p1: "Vertex", p2: "Vertex", p3: "Vertex"
    ) -> "ConstructionCircle":
        """Creates a circle from three points, all points have to be compatible
        to :class:`Vec2` class.
        """
        _p1 = Vec2(p1)
        _p2 = Vec2(p2)
        _p3 = Vec2(p3)
        ray1 = ConstructionRay(_p1, _p2)
        ray2 = ConstructionRay(_p1, _p3)
        center_ray1 = ray1.orthogonal(_p1.lerp(_p2))
        center_ray2 = ray2.orthogonal(_p1.lerp(_p3))
        center = center_ray1.intersect(center_ray2)
        return ConstructionCircle(center, center.distance(_p1))

    @property
    def bounding_box(self) -> "BoundingBox2d":
        """2D bounding box of circle as  :class:`BoundingBox2d` object."""
        rvec = Vec2((self.radius, self.radius))
        return BoundingBox2d((self.center - rvec, self.center + rvec))

    def translate(self, dx: float, dy: float) -> None:
        """Move circle about `dx` in x-axis and about `dy` in y-axis.

        Args:
            dx: translation in x-axis
            dy: translation in y-axis

        """
        self.center += Vec2((dx, dy))

    def point_at(self, angle: float) -> Vec2:
        """Returns point on circle at `angle` as :class:`Vec2` object.

        Args:
            angle: angle in radians

        """
        return self.center + Vec2.from_angle(angle, self.radius)

    def inside(self, point: "Vertex") -> bool:
        """Returns ``True`` if `point` is inside circle."""
        return self.radius >= self.center.distance(Vec2(point))

    def tangent(self, angle: float) -> ConstructionRay:
        """Returns tangent to circle at `angle` as :class:`ConstructionRay`
        object.

        Args:
            angle: angle in radians

        """
        point_on_circle = self.point_at(angle)
        ray = ConstructionRay(self.center, point_on_circle)
        return ray.orthogonal(point_on_circle)

    def intersect_ray(
        self, ray: ConstructionRay, abs_tol: float = 1e-10
    ) -> Sequence[Vec2]:
        """Returns intersection points of circle and `ray` as sequence of
        :class:`Vec2` objects.

        Args:
            ray: intersection ray
            abs_tol: absolute tolerance for tests (e.g. test for tangents)

        Returns:
            tuple of :class:`Vec2` objects

            =========== ==================================
            tuple size  Description
            =========== ==================================
            0           no intersection
            1           ray is a tangent to circle
            2           ray intersects with the circle
            =========== ==================================

        """
        ortho_ray = ray.orthogonal(self.center)
        intersection_point = ray.intersect(ortho_ray)
        dist = self.center.distance(intersection_point)
        result = []
        # Intersect in two points:
        if dist < self.radius:
            # Ray goes through center point:
            if math.isclose(dist, 0.0, abs_tol=abs_tol):
                angle = ortho_ray.angle
                alpha = HALF_PI
            else:
                # The exact direction of angle (all 4 quadrants Q1-Q4) is
                # important: ortho_ray.angle is only correct at the center point
                angle = (intersection_point - self.center).angle
                alpha = math.acos(
                    intersection_point.distance(self.center) / self.radius
                )
            result.append(self.point_at(angle + alpha))
            result.append(self.point_at(angle - alpha))
        # Ray is a tangent of the circle:
        elif math.isclose(dist, self.radius, abs_tol=abs_tol):
            result.append(intersection_point)
        # else: No intersection
        return tuple(result)

    def intersect_circle(
        self, other: "ConstructionCircle", abs_tol: float = 1e-10
    ) -> Sequence[Vec2]:
        """Returns intersection points of two circles as sequence of
        :class:`Vec2` objects.

        Args:
            other: intersection circle
            abs_tol: absolute tolerance for tests

        Returns:
            tuple of :class:`Vec2` objects

            =========== ==================================
            tuple size  Description
            =========== ==================================
            0           no intersection
            1           circle touches the `other` circle at one point
            2           circle intersects with the `other` circle
            =========== ==================================

        """
        r1 = self.radius
        r2 = other.radius
        d = self.center.distance(other.center)
        d_max = r1 + r2
        d_min = math.fabs(r1 - r2)
        result = []
        if d_min <= d <= d_max:
            angle = (other.center - self.center).angle
            # Circles touches at one point:
            if math.isclose(d, d_max, abs_tol=abs_tol) or math.isclose(
                d, d_min, abs_tol=abs_tol
            ):
                result.append(self.point_at(angle))
            else:  # Circles intersect in two points:
                # Law of Cosines:
                alpha = math.acos(
                    (r2 ** 2 - r1 ** 2 - d ** 2) / (-2.0 * r1 * d)
                )
                result.append(self.point_at(angle + alpha))
                result.append(self.point_at(angle - alpha))
        return tuple(result)



# Copyright (c) 2018-2021 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterator, cast, Optional, Tuple
from ezdxf.lldxf.const import DXFValueError

if TYPE_CHECKING:
    from ezdxf.eztypes import (
        DXFObject,
        Dictionary,
        Drawing,
        ObjectsSection,
        EntityDB,
    )


class ObjectCollection:
    def __init__(
        self,
        doc: "Drawing",
        dict_name: str = "ACAD_MATERIAL",
        object_type: str = "MATERIAL",
    ):
        self.doc: "Drawing" = doc
        self.object_type: str = object_type
        self.object_dict: "Dictionary" = doc.rootdict.get_required_dict(
            dict_name
        )

    @property
    def objects(self) -> "ObjectsSection":
        return self.doc.objects

    @property
    def entitydb(self) -> "EntityDB":
        return self.doc.entitydb

    def __iter__(self) -> Iterator[Tuple[str, "DXFObject"]]:
        return self.object_dict.items()

    def __len__(self) -> int:
        return len(self.object_dict)

    def __contains__(self, name: str) -> bool:
        return name in self.object_dict

    def __getitem__(self, name: str) -> "DXFObject":
        return cast("DXFObject", self.object_dict.__getitem__(name))

    def get(
        self, name: str, default: "DXFObject" = None
    ) -> Optional["DXFObject"]:
        """Get object by name.

        Args:
            name: object name as string
            default: default value

        """
        return self.object_dict.get(name, default)  # type: ignore

    def new(self, name: str) -> "DXFObject":
        """Create a new object of type `self.object_type` and store its handle
        in the object manager dictionary.

        Args:
            name: name of new object as string

        Returns:
            new object of type `self.object_type`

        Raises:
            DXFValueError: if object name already exist

        (internal API)

        """
        if name in self.object_dict:
            raise DXFValueError(
                f"{self.object_type} entry {name} already exists."
            )
        return self._new(name, dxfattribs={"name": name})

    def _new(self, name: str, dxfattribs: dict) -> "DXFObject":
        owner = self.object_dict.dxf.handle
        dxfattribs["owner"] = owner
        obj = self.objects.add_dxf_object_with_reactor(
            self.object_type, dxfattribs=dxfattribs
        )
        self.object_dict.add(name, obj)
        return cast("DXFObject", obj)

    def delete(self, name: str) -> None:
        obj = self.object_dict.get(name)
        if obj is not None:
            obj = cast("DXFObject", obj)
            self.object_dict.discard(name)
            self.objects.delete_entity(obj)

    def clear(self) -> None:
        """Delete all entries."""
        self.object_dict.clear()



# Copyright (c) 2011-2021, Manfred Moitzi
# License: MIT License
"""
Tags
----

A list of :class:`~ezdxf.lldxf.types.DXFTag`, inherits from Python standard list.
Unlike the statement in the DXF Reference "Do not write programs that rely on
the order given here", tag order is sometimes essential and some group codes
may appear multiples times in one entity. At the worst case
(:class:`~ezdxf.entities.material.Material`: normal map shares group codes with
diffuse map) using same group codes with different meanings.

"""
from typing import Iterable, Tuple, Iterator, Any

from .const import DXFStructureError, DXFValueError, STRUCTURE_MARKER
from .types import DXFTag, EMBEDDED_OBJ_MARKER, EMBEDDED_OBJ_STR, dxftag
from .tagger import internal_tag_compiler

COMMENT_CODE = 999


class Tags(list):
    """Collection of :class:`~ezdxf.lldxf.types.DXFTag` as flat list.
    Low level tag container, only required for advanced stuff.

    """

    @classmethod
    def from_text(cls, text: str) -> "Tags":
        """Constructor from DXF string."""
        return cls(internal_tag_compiler(text))

    def __copy__(self) -> "Tags":
        return self.__class__(tag.clone() for tag in self)

    clone = __copy__

    def get_handle(self) -> str:
        """Get DXF handle. Raises :class:`DXFValueError` if handle not exist.

        Returns:
            handle as plain hex string like ``'FF00'``

        Raises:
            DXFValueError: no handle found

        """
        try:
            code, handle = self[1]  # fast path  for most common cases
        except IndexError:
            raise DXFValueError("No handle found.")

        if code == 5 or code == 105:
            return handle

        for code, handle in self:
            if code in (5, 105):
                return handle
        raise DXFValueError("No handle found.")

    def replace_handle(self, new_handle: str) -> None:
        """Replace existing handle.

        Args:
            new_handle: new handle as plain hex string e.g. ``'FF00'``

        """
        for index, tag in enumerate(self):
            if tag.code in (5, 105):
                self[index] = DXFTag(tag.code, new_handle)
                return

    def dxftype(self) -> str:
        """Returns DXF type of entity, e.g. ``'LINE'``."""
        return self[0].value

    def has_tag(self, code: int) -> bool:
        """Returns ``True`` if a :class:`~ezdxf.lldxf.types.DXFTag` with given
        group `code` is present.

        Args:
            code: group code as int

        """
        return any(tag.code == code for tag in self)

    def get_first_value(self, code: int, default=DXFValueError) -> Any:
        """Returns value of first :class:`~ezdxf.lldxf.types.DXFTag` with given
        group code or default if `default` != :class:`DXFValueError`, else
        raises :class:`DXFValueError`.

        Args:
            code: group code as int
            default: return value for default case or raises :class:`DXFValueError`

        """
        for tag in self:
            if tag.code == code:
                return tag.value
        if default is DXFValueError:
            raise DXFValueError(code)
        else:
            return default

    def get_first_tag(self, code: int, default=DXFValueError) -> DXFTag:
        """Returns first :class:`~ezdxf.lldxf.types.DXFTag` with given group
        code or `default`, if `default` != :class:`DXFValueError`, else raises
        :class:`DXFValueError`.

        Args:
            code: group code as int
            default: return value for default case or raises :class:`DXFValueError`

        """
        for tag in self:
            if tag.code == code:
                return tag
        if default is DXFValueError:
            raise DXFValueError(code)
        else:
            return default  # type: ignore

    def find_all(self, code: int) -> "Tags":
        """Returns a list of :class:`~ezdxf.lldxf.types.DXFTag` with given
        group code.

        Args:
            code: group code as int

        """
        return self.__class__(tag for tag in self if tag.code == code)

    def tag_index(self, code: int, start: int = 0, end: int = None) -> int:
        """Return index of first :class:`~ezdxf.lldxf.types.DXFTag` with given
        group code.

        Args:
            code: group code as int
            start: start index as int
            end: end index as int, ``None`` for end index = ``len(self)``

        """
        if end is None:
            end = len(self)
        index = start
        while index < end:
            if self[index].code == code:
                return index
            index += 1
        raise DXFValueError(code)

    def update(self, tag: DXFTag) -> None:
        """Update first existing tag with same group code as `tag`, raises
        :class:`DXFValueError` if tag not exist.

        """
        index = self.tag_index(tag.code)
        self[index] = tag

    def set_first(self, tag: DXFTag) -> None:
        """Update first existing tag with group code ``tag.code`` or append tag."""
        try:
            self.update(tag)
        except DXFValueError:
            self.append(tag)

    def remove_tags(self, codes: Iterable[int]) -> None:
        """Remove all tags inplace with group codes specified in `codes`.

        Args:
            codes: iterable of group codes as int

        """
        self[:] = [tag for tag in self if tag.code not in set(codes)]

    def pop_tags(self, codes: Iterable[int]) -> Iterator[DXFTag]:
        """Pop tags with group codes specified in `codes`.

        Args:
            codes: iterable of group codes

        """
        remaining = []
        codes = set(codes)
        for tag in self:
            if tag.code in codes:
                yield tag
            else:
                remaining.append(tag)
        self[:] = remaining

    def remove_tags_except(self, codes: Iterable[int]) -> None:
        """Remove all tags inplace except those with group codes specified in
        `codes`.

        Args:
            codes: iterable of group codes

        """
        self[:] = [tag for tag in self if tag.code in set(codes)]

    def filter(self, codes: Iterable[int]) -> Iterator[DXFTag]:
        """Iterate and filter tags by group `codes`.

        Args:
            codes: group codes to filter

        """
        return (tag for tag in self if tag.code not in set(codes))

    def collect_consecutive_tags(
        self, codes: Iterable[int], start: int = 0, end: int = None
    ) -> "Tags":
        """Collect all consecutive tags with group code in `codes`, `start` and
        `end` delimits the search range. A tag code not in codes ends the
        process.

        Args:
            codes: iterable of group codes
            start: start index as int
            end: end index as int, ``None`` for end index = ``len(self)``

        Returns:
            collected tags as :class:`Tags`

        """
        codes = frozenset(codes)
        index = int(start)
        if end is None:
            end = len(self)
        bag = self.__class__()

        while index < end:
            tag = self[index]
            if tag.code in codes:
                bag.append(tag)
                index += 1
            else:
                break
        return bag

    def has_embedded_objects(self) -> bool:
        for tag in self:
            if (
                tag.code == EMBEDDED_OBJ_MARKER
                and tag.value == EMBEDDED_OBJ_STR
            ):
                return True
        return False

    @classmethod
    def strip(cls, tags: "Tags", codes: Iterable[int]) -> "Tags":
        """Constructor from `tags`, strips all tags with group codes in `codes`
        from tags.

        Args:
            tags: iterable of :class:`~ezdxf.lldxf.types.DXFTag`
            codes: iterable of group codes as int

        """
        return cls((tag for tag in tags if tag.code not in frozenset(codes)))


def text2tags(text: str) -> Tags:
    return Tags.from_text(text)


def group_tags(
    tags: Iterable[DXFTag], splitcode: int = STRUCTURE_MARKER
) -> Iterable[Tags]:
    """Group of tags starts with a SplitTag and ends before the next SplitTag.
    A SplitTag is a tag with code == splitcode, like (0, 'SECTION') for
    splitcode == 0.

    Args:
        tags: iterable of :class:`DXFTag`
        splitcode: group code of split tag

    """

    # first do nothing, skip tags in front of the first split tag
    def append(tag):
        pass

    group = None
    for tag in tags:
        if tag.code == splitcode:
            if group is not None:
                yield group
            group = Tags([tag])
            append = group.append  # redefine append: add tags to this group
        else:
            append(tag)
    if group is not None:
        yield group


def text_to_multi_tags(
    text: str, code: int = 303, size: int = 255, line_ending: str = "^J"
) -> Tags:
    text = "".join(text).replace("\n", line_ending)

    def chop():
        start = 0
        end = size
        while start < len(text):
            yield text[start:end]
            start = end
            end += size

    return Tags(DXFTag(code, part) for part in chop())


def multi_tags_to_text(tags: Tags, line_ending: str = "^J") -> str:
    return "".join(tag.value for tag in tags).replace(line_ending, "\n")


OPEN_LIST = (1002, "{")
CLOSE_LIST = (1002, "}")


def xdata_list(name: str, xdata_tags: Iterable) -> Tags:
    tags = Tags()
    if name:
        tags.append((1000, name))
    tags.append(OPEN_LIST)
    tags.extend(xdata_tags)
    tags.append(CLOSE_LIST)
    return tags


def remove_named_list_from_xdata(name: str, tags: Tags) -> Tags:
    start, end = get_start_and_end_of_named_list_in_xdata(name, tags)
    del tags[start:end]
    return tags


def get_named_list_from_xdata(name: str, tags: Tags) -> Tags:
    start, end = get_start_and_end_of_named_list_in_xdata(name, tags)
    return Tags(tags[start:end])


class NotFoundException(Exception):
    pass


def get_start_and_end_of_named_list_in_xdata(
    name: str, tags: Tags
) -> Tuple[int, int]:
    start = None
    end = None
    level = 0
    for index in range(len(tags)):
        tag = tags[index]

        if start is None and tag == (1000, name):
            next_tag = tags[index + 1]
            if next_tag == OPEN_LIST:
                start = index
                continue
        if start is not None:
            if tag == OPEN_LIST:
                level += 1
            elif tag == CLOSE_LIST:
                level -= 1
            if level == 0:
                end = index
                break

    if start is None:
        raise NotFoundException
    if end is None:
        raise DXFStructureError(
            'Invalid XDATA structure: missing  (1002, "}").'
        )
    return start, end + 1


def find_begin_and_end_of_encoded_xdata_tags(
    name: str, tags: Tags
) -> Tuple[int, int]:
    """Find encoded XDATA tags, surrounded by group code 1000 tags
    name_BEGIN and name_END (e.g. MTEXT column specification).

    Raises:
        NotFoundError: tag group not found
        DXFStructureError: missing begin- or end tag

    """
    begin_name = name + "_BEGIN"
    end_name = name + "_END"
    start = None
    end = None
    for index, (code, value) in enumerate(tags):
        if code == 1000:
            if value == begin_name:
                start = index
            elif value == end_name:
                end = index + 1
                break
    if start is None:
        if end is not None:  # end tag without begin tag!
            raise DXFStructureError(
                f"Invalid XDATA structure: missing begin tag (1000, {begin_name})."
            )
        raise NotFoundException
    if end is None:
        raise DXFStructureError(
            f"Invalid XDATA structure: missing end tag (1000, {end_name})."
        )
    return start, end


def binary_data_to_dxf_tags(
    data: bytes,
    length_group_code: int = 160,
    value_group_code: int = 310,
    value_size=127,
) -> Tags:
    """Convert binary data to DXF tags."""
    tags = Tags()
    length = len(data)
    tags.append(dxftag(length_group_code, length))
    index = 0
    while index < length:
        chunk = data[index : index + value_size]
        tags.append(dxftag(value_group_code, chunk))
        index += value_size
    return tags



# Copyright (c) 2021 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterable, Sequence, Type, Optional
import math

# The pure Python implementation can't import from ._ctypes or ezdxf.math!
from ._vector import Vec3, Vec2
from ._matrix44 import Matrix44

if TYPE_CHECKING:
    from ezdxf.eztypes import Vertex, AnyVec

__all__ = ["Bezier3P"]


def check_if_in_valid_range(t: float) -> None:
    if not (0 <= t <= 1.0):
        raise ValueError("t not in range [0 to 1]")


class Bezier3P:
    """Implements an optimized quadratic `Bézier curve`_ for exact 3 control
    points.

    Special behavior:

        - 2D control points in, returns 2D results as :class:`~ezdxf.math.Vec2` objects
        - 3D control points in, returns 3D results as :class:`~ezdxf.math.Vec3` objects
        - Object is immutable.

    Args:
        defpoints: iterable of definition points as :class:`Vec2` or
            :class:`Vec3` compatible objects.

    """

    def __init__(self, defpoints: Sequence["Vertex"]):
        if len(defpoints) == 3:
            is3d = any(len(p) > 2 for p in defpoints)
            vector_class: Type["AnyVec"] = Vec3 if is3d else Vec2
            self._control_points: Sequence["AnyVec"] = vector_class.tuple(
                defpoints
            )
        else:
            raise ValueError("Three control points required.")

    @property
    def control_points(self) -> Sequence["AnyVec"]:
        """Control points as tuple of :class:`~ezdxf.math.Vec3` or
        :class:`~ezdxf.math.Vec2` objects.
        """
        return self._control_points

    def tangent(self, t: float) -> "AnyVec":
        """Returns direction vector of tangent for location `t` at the
        Bèzier-curve.

        Args:
            t: curve position in the range ``[0, 1]``

        """
        check_if_in_valid_range(t)
        return self._get_curve_tangent(t)

    def point(self, t: float) -> "AnyVec":
        """Returns point for location `t`` at the Bèzier-curve.

        Args:
            t: curve position in the range ``[0, 1]``

        """
        check_if_in_valid_range(t)
        return self._get_curve_point(t)

    def approximate(self, segments: int) -> Iterable["AnyVec"]:
        """Approximate `Bézier curve`_ by vertices, yields `segments` + 1
        vertices as ``(x, y[, z])`` tuples.

        Args:
            segments: count of segments for approximation

        """
        if segments < 1:
            raise ValueError(segments)
        delta_t: float = 1.0 / segments
        yield self._control_points[0]
        for segment in range(1, segments):
            yield self._get_curve_point(delta_t * segment)
        yield self._control_points[2]

    def approximated_length(self, segments: int = 128) -> float:
        """Returns estimated length of Bèzier-curve as approximation by line
        `segments`.
        """
        length: float = 0.0
        prev_point: Optional["AnyVec"] = None
        for point in self.approximate(segments):
            if prev_point is not None:
                length += prev_point.distance(point)
            prev_point = point
        return length

    def flattening(
        self, distance: float, segments: int = 4
    ) -> Iterable["AnyVec"]:
        """Adaptive recursive flattening. The argument `segments` is the
        minimum count of approximation segments, if the distance from the center
        of the approximation segment to the curve is bigger than `distance` the
        segment will be subdivided.

        Args:
            distance: maximum distance from the center of the quadratic (C2)
                curve to the center of the linear (C1) curve between two
                approximation points to determine if a segment should be
                subdivided.
            segments: minimum segment count

        """

        def subdiv(
            start_point: "AnyVec",
            end_point: "AnyVec",
            start_t: float,
            end_t: float,
        ) -> Iterable["AnyVec"]:
            mid_t: float = (start_t + end_t) * 0.5
            mid_point: "AnyVec" = self._get_curve_point(mid_t)
            chk_point: "AnyVec" = start_point.lerp(end_point)
            # center point point is faster than projecting mid point onto
            # vector start -> end:
            if chk_point.distance(mid_point) < distance:
                yield end_point
            else:
                yield from subdiv(start_point, mid_point, start_t, mid_t)
                yield from subdiv(mid_point, end_point, mid_t, end_t)

        dt: float = 1.0 / segments
        t0: float = 0.0
        t1: float
        start_point: "AnyVec" = self._control_points[0]
        end_point: "AnyVec"
        yield start_point
        while t0 < 1.0:
            t1 = t0 + dt
            if math.isclose(t1, 1.0):
                end_point = self._control_points[2]
                t1 = 1.0
            else:
                end_point = self._get_curve_point(t1)
            yield from subdiv(start_point, end_point, t0, t1)
            t0 = t1
            start_point = end_point

    def _get_curve_point(self, t: float) -> "AnyVec":
        p0, p1, p2 = self._control_points
        _1_minus_t = 1.0 - t
        a = _1_minus_t * _1_minus_t
        b = 2.0 * t * _1_minus_t
        c = t * t
        return p0 * a + p1 * b + p2 * c

    def _get_curve_tangent(self, t: float) -> "AnyVec":
        p0, p1, p2 = self._control_points
        a = -2.0 * (1.0 - t)
        b = 2.0 - 4.0 * t
        c = 2.0 * t
        return p0 * a + p1 * b + p2 * c

    def reverse(self) -> "Bezier3P":
        """Returns a new Bèzier-curve with reversed control point order."""
        return Bezier3P(list(reversed(self.control_points)))

    def transform(self, m: Matrix44) -> "Bezier3P":
        """General transformation interface, returns a new :class:`Bezier3P`
        curve and it is always a 3D curve.

        Args:
             m: 4x4 transformation matrix (:class:`ezdxf.math.Matrix44`)

        """
        defpoints: Iterable["AnyVec"]
        if len(self._control_points[0]) == 2:
            defpoints = Vec3.generate(self._control_points)
        else:
            defpoints = self._control_points
        return Bezier3P(tuple(m.transform_vertices(defpoints)))



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterable, Sequence, cast, List
import array
import copy
from itertools import chain
from ezdxf.audit import AuditError
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.const import (
    SUBCLASS_MARKER,
    DXF2000,
    DXFValueError,
    DXFStructureError,
)
from ezdxf.lldxf.packedtags import VertexArray, Tags
from ezdxf.math import (
    Vec3,
    Matrix44,
    ConstructionEllipse,
    Z_AXIS,
    NULLVEC,
    OCS,
    uniform_knot_vector,
    open_uniform_knot_vector,
    BSpline,
    required_knot_values,
    required_fit_points,
    required_control_points,
    fit_points_to_cad_cv,
)
from .dxfentity import base_class, SubclassProcessor, DXFEntity
from .dxfgfx import DXFGraphic, acdb_entity
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, DXFNamespace, Vertex, Auditor, Ellipse

__all__ = ["Spline"]

# From the Autodesk ObjectARX reference:
# Objects of the AcDbSpline class use an embedded gelib object to maintain the
# actual spline information.
#
# Book recommendations:
#
#  - "Curves and Surfaces for CAGD" by Gerald Farin
#  - "Mathematical Elements for Computer Graphics"
#    by David Rogers and Alan Adams
#  - "An Introduction To Splines For Use In Computer Graphics & Geometric Modeling"
#    by Richard H. Bartels, John C. Beatty, and Brian A Barsky
#
# http://help.autodesk.com/view/OARX/2018/ENU/?guid=OREF-AcDbSpline__setFitData_AcGePoint3dArray__AcGeVector3d__AcGeVector3d__AcGe__KnotParameterization_int_double
# Construction of a AcDbSpline entity from fit points:
# degree has no effect. A spline with degree=3 is always constructed when
# interpolating a series of fit points.

acdb_spline = DefSubclass(
    "AcDbSpline",
    {
        # Spline flags:
        # 1 = Closed spline
        # 2 = Periodic spline
        # 4 = Rational spline
        # 8 = Planar
        # 16 = Linear (planar bit is also set)
        "flags": DXFAttr(70, default=0),
        # degree: The degree can't be higher than 11 according to the Autodesk
        # ObjectARX reference.
        "degree": DXFAttr(71, default=3, validator=validator.is_positive),
        "n_knots": DXFAttr(72, xtype=XType.callback, getter="knot_count"),
        "n_control_points": DXFAttr(
            73, xtype=XType.callback, getter="control_point_count"
        ),
        "n_fit_points": DXFAttr(
            74, xtype=XType.callback, getter="fit_point_count"
        ),
        "knot_tolerance": DXFAttr(42, default=1e-10, optional=True),
        "control_point_tolerance": DXFAttr(43, default=1e-10, optional=True),
        "fit_tolerance": DXFAttr(44, default=1e-10, optional=True),
        # Start- and end tangents should be normalized, but CAD applications do not
        # crash if they are not normalized.
        "start_tangent": DXFAttr(
            12,
            xtype=XType.point3d,
            optional=True,
            validator=validator.is_not_null_vector,
        ),
        "end_tangent": DXFAttr(
            13,
            xtype=XType.point3d,
            optional=True,
            validator=validator.is_not_null_vector,
        ),
        # Extrusion is the normal vector (omitted if the spline is non-planar)
        "extrusion": DXFAttr(
            210,
            xtype=XType.point3d,
            default=Z_AXIS,
            optional=True,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
        # 10: Control points (in WCS); one entry per control point
        # 11: Fit points (in WCS); one entry per fit point
        # 40: Knot value (one entry per knot)
        # 41: Weight (if not 1); with multiple group pairs, they are present if all
        #     are not 1
    },
)
acdb_spline_group_codes = group_code_mapping(acdb_spline)


class SplineData:
    def __init__(self, spline: "Spline"):
        self.fit_points = spline.fit_points
        self.control_points = spline.control_points
        self.knots = spline.knots
        self.weights = spline.weights


REMOVE_CODES = {10, 11, 40, 41, 72, 73, 74}

Vertices = List[Sequence[float]]


@register_entity
class Spline(DXFGraphic):
    """DXF SPLINE entity"""

    DXFTYPE = "SPLINE"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_spline)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000
    CLOSED = 1  # closed b-spline
    PERIODIC = 2  # uniform b-spline
    RATIONAL = 4  # rational b-spline
    PLANAR = 8  # all spline points in a plane, don't read or set this bit, just ignore like AutoCAD
    LINEAR = 16  # always set with PLANAR, don't read or set this bit, just ignore like AutoCAD

    def __init__(self):
        super().__init__()
        self.fit_points = VertexArray()
        self.control_points = VertexArray()
        self.knots = []
        self.weights = []

    def _copy_data(self, entity: DXFEntity) -> None:
        """Copy data: control_points, fit_points, weights, knot_values."""
        assert isinstance(entity, Spline)
        entity._control_points = copy.deepcopy(self._control_points)
        entity._fit_points = copy.deepcopy(self._fit_points)
        entity._knots = copy.deepcopy(self._knots)
        entity._weights = copy.deepcopy(self._weights)

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor:
            tags = processor.subclass_by_index(2)
            if tags:
                tags = Tags(self.load_spline_data(tags))
                processor.fast_load_dxfattribs(
                    dxf, acdb_spline_group_codes, subclass=tags, recover=True
                )
            else:
                raise DXFStructureError(
                    f"missing 'AcDbSpline' subclass in SPLINE(#{dxf.handle})"
                )
        return dxf

    def load_spline_data(self, tags) -> Iterable:
        """Load and set spline data (fit points, control points, weights,
        knots) and remove invalid start- and end tangents.
        Yields the remaining unprocessed tags.
        """
        control_points = []
        fit_points = []
        knots = []
        weights = []
        for tag in tags:
            code, value = tag
            if code == 10:
                control_points.append(value)
            elif code == 11:
                fit_points.append(value)
            elif code == 40:
                knots.append(value)
            elif code == 41:
                weights.append(value)
            elif code in (12, 13) and NULLVEC.isclose(value):
                # Tangent values equal to (0, 0, 0) are invalid and ignored at
                # the loading stage!
                pass
            else:
                yield tag
        self.control_points = control_points
        self.fit_points = fit_points
        self.knots = knots
        self.weights = weights

    def export_entity(self, tagwriter: "TagWriter") -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_spline.name)
        self.dxf.export_dxf_attribs(tagwriter, ["extrusion", "flags", "degree"])
        tagwriter.write_tag2(72, self.knot_count())
        tagwriter.write_tag2(73, self.control_point_count())
        tagwriter.write_tag2(74, self.fit_point_count())
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "knot_tolerance",
                "control_point_tolerance",
                "fit_tolerance",
                "start_tangent",
                "end_tangent",
            ],
        )

        self.export_spline_data(tagwriter)

    def export_spline_data(self, tagwriter: "TagWriter"):
        for value in self._knots:
            tagwriter.write_tag2(40, value)

        if len(self._weights):
            for value in self._weights:
                tagwriter.write_tag2(41, value)

        self._control_points.export_dxf(tagwriter, code=10)  # type: ignore
        self._fit_points.export_dxf(tagwriter, code=11)  # type: ignore

    @property
    def closed(self) -> bool:
        """``True`` if spline is closed. A closed spline has a connection from
        the last control point to the first control point. (read/write)
        """
        return self.get_flag_state(self.CLOSED, name="flags")

    @closed.setter
    def closed(self, status: bool) -> None:
        self.set_flag_state(self.CLOSED, state=status, name="flags")

    @property
    def knots(self) -> List[float]:
        """Knot values as :code:`array.array('d')`."""
        return self._knots

    @knots.setter
    def knots(self, values: Iterable[float]) -> None:
        self._knots: List[float] = cast(List[float], array.array("d", values))

    # DXF callback attribute Spline.dxf.n_knots
    def knot_count(self) -> int:
        """Count of knot values."""
        return len(self._knots)

    @property
    def weights(self) -> List[float]:
        """Control point weights as :code:`array.array('d')`."""
        return self._weights

    @weights.setter
    def weights(self, values: Iterable[float]) -> None:
        self._weights: List[float] = cast(List[float], array.array("d", values))

    @property
    def control_points(self) -> Vertices:
        """:class:`~ezdxf.lldxf.packedtags.VertexArray` of control points in
        :ref:`WCS`.
        """
        return self._control_points

    @control_points.setter
    def control_points(self, points: Iterable["Vertex"]) -> None:
        self._control_points: Vertices = cast(
            Vertices, VertexArray(chain.from_iterable(Vec3.generate(points)))
        )

    # DXF callback attribute Spline.dxf.n_control_points
    def control_point_count(self) -> int:
        """Count of control points."""
        return len(self.control_points)

    @property
    def fit_points(self) -> Vertices:
        """:class:`~ezdxf.lldxf.packedtags.VertexArray` of fit points in
        :ref:`WCS`.
        """
        return self._fit_points

    @fit_points.setter
    def fit_points(self, points: Iterable["Vertex"]) -> None:
        self._fit_points: Vertices = cast(
            Vertices,
            VertexArray(chain.from_iterable(Vec3.generate(points))),
        )

    # DXF callback attribute Spline.dxf.n_fit_points
    def fit_point_count(self) -> int:
        """Count of fit points."""
        return len(self.fit_points)

    def construction_tool(self) -> BSpline:
        """Returns the construction tool :class:`ezdxf.math.BSpline`."""
        if self.control_point_count():
            weights = self.weights if len(self.weights) else None
            knots = self.knots if len(self.knots) else None
            return BSpline(
                control_points=self.control_points,
                order=self.dxf.degree + 1,
                knots=knots,
                weights=weights,
            )
        elif self.fit_point_count():
            tangents = None
            if self.dxf.hasattr("start_tangent") and self.dxf.hasattr(
                "end_tangent"
            ):
                tangents = [self.dxf.start_tangent, self.dxf.end_tangent]
            # SPLINE from fit points has always a degree of 3!
            return fit_points_to_cad_cv(
                self.fit_points,
                tangents=tangents,
            )
        else:
            raise ValueError(
                "Construction tool requires control- or fit points."
            )

    def apply_construction_tool(self, s) -> "Spline":
        """Apply SPLINE data from a :class:`~ezdxf.math.BSpline` construction
        tool or from a :class:`geomdl.BSpline.Curve` object.

        """
        try:
            self.control_points = s.control_points
        except AttributeError:  # maybe a geomdl.BSpline.Curve class
            s = BSpline.from_nurbs_python_curve(s)
            self.control_points = s.control_points

        self.dxf.degree = s.degree
        self.fit_points = []  # remove fit points
        self.knots = s.knots()
        self.weights = s.weights()
        self.set_flag_state(Spline.RATIONAL, state=bool(len(self.weights)))
        return self  # floating interface

    def flattening(self, distance: float, segments: int = 4) -> Iterable[Vec3]:
        """Adaptive recursive flattening. The argument `segments` is the
        minimum count of approximation segments between two knots, if the
        distance from the center of the approximation segment to the curve is
        bigger than `distance` the segment will be subdivided.

        Args:
            distance: maximum distance from the projected curve point onto the
                segment chord.
            segments: minimum segment count between two knots

        .. versionadded:: 0.15

        """
        return self.construction_tool().flattening(distance, segments)

    @classmethod
    def from_arc(cls, entity: "DXFGraphic") -> "Spline":
        """Create a new SPLINE entity from a CIRCLE, ARC or ELLIPSE entity.

        The new SPLINE entity has no owner, no handle, is not stored in
        the entity database nor assigned to any layout!

        """
        dxftype = entity.dxftype()
        if dxftype == "ELLIPSE":
            ellipse = cast("Ellipse", entity).construction_tool()
        elif dxftype == "CIRCLE":
            ellipse = ConstructionEllipse.from_arc(
                center=entity.dxf.get("center", NULLVEC),
                radius=abs(entity.dxf.get("radius", 1.0)),
                extrusion=entity.dxf.get("extrusion", Z_AXIS),
            )
        elif dxftype == "ARC":
            ellipse = ConstructionEllipse.from_arc(
                center=entity.dxf.get("center", NULLVEC),
                radius=abs(entity.dxf.get("radius", 1.0)),
                extrusion=entity.dxf.get("extrusion", Z_AXIS),
                start_angle=entity.dxf.get("start_angle", 0),
                end_angle=entity.dxf.get("end_angle", 360),
            )
        else:
            raise TypeError("CIRCLE, ARC or ELLIPSE entity required.")

        spline = Spline.new(
            dxfattribs=entity.graphic_properties(), doc=entity.doc
        )
        s = BSpline.from_ellipse(ellipse)
        spline.dxf.degree = s.degree
        spline.dxf.flags = Spline.RATIONAL
        spline.control_points = s.control_points  # type: ignore
        spline.knots = s.knots()  # type: ignore
        spline.weights = s.weights()  # type: ignore
        return spline

    def set_open_uniform(
        self, control_points: Sequence["Vertex"], degree: int = 3
    ) -> None:
        """Open B-spline with uniform knot vector, start and end at your first
        and last control points.

        """
        self.dxf.flags = 0
        self.dxf.degree = degree
        self.control_points = control_points  # type: ignore
        self.knots = open_uniform_knot_vector(len(control_points), degree + 1)

    def set_uniform(
        self, control_points: Sequence["Vertex"], degree: int = 3
    ) -> None:
        """B-spline with uniform knot vector, does NOT start and end at your
        first and last control points.

        """
        self.dxf.flags = 0
        self.dxf.degree = degree
        self.control_points = control_points  # type: ignore
        self.knots = uniform_knot_vector(len(control_points), degree + 1)

    def set_closed(self, control_points: Sequence["Vertex"], degree=3) -> None:
        """
        Closed B-spline with uniform knot vector, start and end at your first control point.

        """
        self.dxf.flags = self.PERIODIC | self.CLOSED
        self.dxf.degree = degree
        self.control_points = control_points  # type: ignore
        self.control_points.extend(control_points[:degree])
        # AutoDesk Developer Docs:
        # If the spline is periodic, the length of knot vector will be greater
        # than length of the control array by 1, but this does not work with
        # BricsCAD.
        self.knots = uniform_knot_vector(len(self.control_points), degree + 1)

    def set_open_rational(
        self,
        control_points: Sequence["Vertex"],
        weights: Sequence[float],
        degree: int = 3,
    ) -> None:
        """Open rational B-spline with uniform knot vector, start and end at
        your first and last control points, and has additional control
        possibilities by weighting each control point.

        """
        self.set_open_uniform(control_points, degree=degree)
        self.dxf.flags = self.dxf.flags | self.RATIONAL
        if len(weights) != len(self.control_points):
            raise DXFValueError(
                "Control point count must be equal to weights count."
            )
        self.weights = weights  # type: ignore

    def set_uniform_rational(
        self,
        control_points: Sequence["Vertex"],
        weights: Sequence[float],
        degree: int = 3,
    ) -> None:
        """Rational B-spline with uniform knot vector, does NOT start and end
        at your first and last control points, and has additional control
        possibilities by weighting each control point.

        """
        self.set_uniform(control_points, degree=degree)
        self.dxf.flags = self.dxf.flags | self.RATIONAL
        if len(weights) != len(self.control_points):
            raise DXFValueError(
                "Control point count must be equal to weights count."
            )
        self.weights = weights  # type: ignore

    def set_closed_rational(
        self,
        control_points: Sequence["Vertex"],
        weights: Sequence[float],
        degree: int = 3,
    ) -> None:
        """Closed rational B-spline with uniform knot vector, start and end at
        your first control point, and has additional control possibilities by
        weighting each control point.

        """
        self.set_closed(control_points, degree=degree)
        self.dxf.flags = self.dxf.flags | self.RATIONAL
        weights = list(weights)
        weights.extend(weights[:degree])
        if len(weights) != len(self.control_points):
            raise DXFValueError(
                "Control point count must be equal to weights count."
            )
        self.weights = weights

    def transform(self, m: "Matrix44") -> "Spline":
        """Transform the SPLINE entity by transformation matrix `m` inplace."""
        self._control_points.transform(m)  # type: ignore
        self._fit_points.transform(m)  # type: ignore
        # Transform optional attributes if they exist
        dxf = self.dxf
        for name in ("start_tangent", "end_tangent", "extrusion"):
            if dxf.hasattr(name):
                dxf.set(name, m.transform_direction(dxf.get(name)))
        self.post_transform(m)
        return self

    def audit(self, auditor: "Auditor") -> None:
        """Audit the SPLINE entity.

        .. versionadded:: 0.15.1

        """
        super().audit(auditor)
        degree = self.dxf.degree
        name = str(self)

        if degree < 1:
            auditor.fixed_error(
                code=AuditError.INVALID_SPLINE_DEFINITION,
                message=f"Removed {name} with invalid degree: {degree} < 1.",
            )
            auditor.trash(self)
            return

        n_control_points = len(self.control_points)
        n_fit_points = len(self.fit_points)

        if n_control_points == 0 and n_fit_points == 0:
            auditor.fixed_error(
                code=AuditError.INVALID_SPLINE_DEFINITION,
                message=f"Removed {name} without any points (no geometry).",
            )
            auditor.trash(self)
            return

        if n_control_points > 0:
            self._audit_control_points(auditor)
        # Ignore fit points if defined by control points
        elif n_fit_points > 0:
            self._audit_fit_points(auditor)

    def _audit_control_points(self, auditor: "Auditor"):
        name = str(self)
        order = self.dxf.degree + 1
        n_control_points = len(self.control_points)

        # Splines with to few control points can't be processed:
        n_control_points_required = required_control_points(order)
        if n_control_points < n_control_points_required:
            auditor.fixed_error(
                code=AuditError.INVALID_SPLINE_CONTROL_POINT_COUNT,
                message=f"Removed {name} with invalid control point count: "
                f"{n_control_points} < {n_control_points_required}",
            )
            auditor.trash(self)
            return

        n_weights = len(self.weights)
        n_knots = len(self.knots)
        n_knots_required = required_knot_values(n_control_points, order)

        if n_knots < n_knots_required:
            # Can not fix entity: because the knot values are basic
            # values which define the geometry of SPLINE.
            auditor.fixed_error(
                code=AuditError.INVALID_SPLINE_KNOT_VALUE_COUNT,
                message=f"Removed {name} with invalid knot value count: "
                f"{n_knots} < {n_knots_required}",
            )
            auditor.trash(self)
            return

        if n_weights and n_weights != n_control_points:
            # Can not fix entity: because the weights are basic
            # values which define the geometry of SPLINE.
            auditor.fixed_error(
                code=AuditError.INVALID_SPLINE_WEIGHT_COUNT,
                message=f"Removed {name} with invalid weight count: "
                f"{n_weights} != {n_control_points}",
            )
            auditor.trash(self)
            return

    def _audit_fit_points(self, auditor: "Auditor"):
        name = str(self)
        order = self.dxf.degree + 1
        # Assuming end tangents will be estimated if not present,
        # like by ezdxf:
        n_fit_points_required = required_fit_points(order, tangents=True)

        # Splines with to few fit points can't be processed:
        n_fit_points = len(self.fit_points)
        if n_fit_points < n_fit_points_required:
            auditor.fixed_error(
                code=AuditError.INVALID_SPLINE_FIT_POINT_COUNT,
                message=f"Removed {name} with invalid fit point count: "
                f"{n_fit_points} < {n_fit_points_required}",
            )
            auditor.trash(self)
            return

        # Knot values have no meaning for splines defined by fit points:
        if len(self.knots):
            auditor.fixed_error(
                code=AuditError.INVALID_SPLINE_KNOT_VALUE_COUNT,
                message=f"Removed unused knot values for {name} "
                f"defined by fit points.",
            )
            self.knots = []

        # Weights have no meaning for splines defined by fit points:
        if len(self.weights):
            auditor.fixed_error(
                code=AuditError.INVALID_SPLINE_WEIGHT_COUNT,
                message=f"Removed unused weights for {name} "
                f"defined by fit points.",
            )
            self.weights = []

    def ocs(self) -> OCS:
        # WCS entity which supports the "extrusion" attribute in a
        # different way!
        return OCS()



# Copyright (c) 2015-2021, Manfred Moitzi
# License: MIT License
from typing import Tuple, Any, Iterable
from uuid import uuid4
import functools
import html
from .juliandate import juliandate, calendardate
from .binarydata import hex_strings_to_bytes, bytes_to_hexstr

escape = functools.partial(html.escape, quote=True)


def float2transparency(value: float) -> int:
    """
    Returns DXF transparency value as integer in the range from ``0`` to ``255``, where ``0`` is 100% transparent
    and ``255`` is opaque.

    Args:
        value: transparency value as float in the range from ``0`` to ``1``, where ``0`` is opaque
               and ``1`` is 100% transparency.

    """
    return int((1.0 - float(value)) * 255) | 0x02000000


def transparency2float(value: int) -> float:
    """
    Returns transparency value as float from ``0`` to ``1``, ``0`` for no transparency (opaque) and ``1``
    for 100% transparency.

    Args:
        value: DXF integer transparency value, ``0`` for 100% transparency and ``255`` for opaque

    """
    # 255 -> 0.
    # 0 -> 1.
    return 1.0 - float(int(value) & 0xFF) / 255.0


def set_flag_state(flags: int, flag: int, state: bool = True) -> int:
    """Set/clear binary `flag` in data `flags`.

    Args:
        flags: data value
        flag: flag to set/clear
        state: ``True`` for setting, ``False`` for clearing

    """
    if state:
        flags = flags | flag
    else:
        flags = flags & ~flag
    return flags


def guid() -> str:
    """Returns a general unique ID, based on :func:`uuid.uuid4`.

    This function creates a GUID for the header variables $VERSIONGUID and
    $FINGERPRINTGUID, which matches the AutoCAD pattern
    ``{XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX}``.

    """
    return "{" + str(uuid4()).upper() + "}"


def take2(iterable: Iterable) -> Iterable[Tuple[Any, Any]]:
    """Iterate `iterable` as 2-tuples.

    :code:`[1, 2, 3, 4, ...] -> (1, 2), (3, 4), ...`

    """
    store = None
    for item in iterable:
        if store is None:
            store = item
        else:
            yield store, item
            store = None


def suppress_zeros(s: str, leading: bool = False, trailing: bool = True):
    """Suppress trailing and/or leading ``0`` of string `s`.

    Args:
         s: data string
         leading: suppress leading ``0``
         trailing: suppress trailing ``0``

    """
    # is anything to do?
    if (not leading) and (not trailing):
        return s

    # if `s` represents zero
    if float(s) == 0.0:
        return "0"

    # preserve sign
    if s[0] in "-+":
        sign = s[0]
        s = s[1:]
    else:
        sign = ""

    # strip zeros
    if leading:
        s = s.lstrip("0")
    if trailing:
        s = s.rstrip("0")

    # remove comma if no decimals follow
    if s[-1] in ".,":
        s = s[:-1]

    return sign + s


def normalize_text_angle(angle: float, fix_upside_down=True) -> float:
    """
    Normalizes text `angle` to the range from 0 to 360 degrees and fixes upside down text angles.

    Args:
        angle: text angle in degrees
        fix_upside_down: rotate upside down text angle about 180 degree

    """
    angle = angle % 360.0  # normalize angle (0 .. 360)
    if fix_upside_down and (90 < angle <= 270):  # flip text orientation
        angle -= 180
        angle = angle % 360.0  # normalize again
    return angle



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING
from ezdxf.lldxf.attributes import DXFAttr, DXFAttributes, DefSubclass
from ezdxf.lldxf import const
from .dxfentity import SubclassProcessor, DXFEntity
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.eztypes import TagWriter, DXFNamespace

__all__ = ["TableHead"]

base_class = DefSubclass(
    None,
    {
        "name": DXFAttr(2),
        "handle": DXFAttr(5),
        "owner": DXFAttr(330),
    },
)

acdb_symbol_table = DefSubclass(
    "AcDbSymbolTable",
    {
        "count": DXFAttr(70, default=0),
    },
)


@register_entity
class TableHead(DXFEntity):
    """The table head structure is only maintained for export and not for
    internal usage, ezdxf ignores an inconsistent table head at runtime.

    """

    DXFTYPE = "TABLE"
    DXFATTRIBS = DXFAttributes(base_class, acdb_symbol_table)

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor:
            dxf.name = processor.base_class.get_first_value(2)
            # Stored max table count is not required:
            dxf.count = 0
        return dxf

    def export_dxf(self, tagwriter: "TagWriter") -> None:
        assert self.dxf.handle, (
            "TABLE needs a handle, maybe loaded from " "DXF R12 without handle!"
        )
        tagwriter.write_tag2(const.STRUCTURE_MARKER, self.DXFTYPE)
        tagwriter.write_tag2(2, self.dxf.name)
        if tagwriter.dxfversion >= const.DXF2000:
            tagwriter.write_tag2(5, self.dxf.handle)
            if self.has_extension_dict:
                self.extension_dict.export_dxf(tagwriter)  # type: ignore
            tagwriter.write_tag2(const.OWNER_CODE, self.dxf.owner)
            tagwriter.write_tag2(const.SUBCLASS_MARKER, acdb_symbol_table.name)
            tagwriter.write_tag2(70, self.dxf.count)
            # There is always one exception:
            if self.dxf.name == "DIMSTYLE":
                tagwriter.write_tag2(const.SUBCLASS_MARKER, "AcDbDimStyleTable")
        else:  # DXF R12
            # TABLE does not need a handle at all
            tagwriter.write_tag2(70, self.dxf.count)



# Copyright (c) 2018-2021, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, List, Union, Optional, Iterable, Dict, Any
import copy
import logging
from collections import namedtuple

from ezdxf.lldxf import const
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    group_code_mapping,
)
from ezdxf.lldxf.tags import Tags
from ezdxf.math import Vec3, NULLVEC, X_AXIS, Y_AXIS, Z_AXIS, Matrix44
from ezdxf import colors
from ezdxf.proxygraphic import ProxyGraphic

from .dxfentity import base_class, SubclassProcessor
from .dxfobj import DXFObject
from .dxfgfx import DXFGraphic, acdb_entity

from .factory import register_entity
from .objectcollection import ObjectCollection

if TYPE_CHECKING:
    from ezdxf.eztypes import (
        TagWriter,
        Drawing,
        DXFNamespace,
        DXFTag,
        DXFEntity,
    )

__all__ = ["MultiLeader", "MLeader", "MLeaderStyle", "MLeaderStyleCollection"]

logger = logging.getLogger("ezdxf")

# DXF Examples:
# "D:\source\dxftest\CADKitSamples\house design for two family with common staircasedwg.dxf"
# "D:\source\dxftest\CADKitSamples\house design.dxf"

# How to render MLEADER: https://atlight.github.io/formats/dxf-leader.html
# DXF reference:
# http://help.autodesk.com/view/OARX/2018/ENU/?guid=GUID-72D20B8C-0F5E-4993-BEB7-0FCF94F32BE0

acdb_mleader = DefSubclass(
    "AcDbMLeader",
    {
        "version": DXFAttr(270, default=2),
        "style_handle": DXFAttr(340),
        # Theory: Take properties from MLEADERSTYLE,
        # except explicit overridden here:
        "property_override_flags": DXFAttr(90),
        # Bit coded flags:
        # 1 << 0 = leader_type
        # 1 << 1 = leader_line_color
        # 1 << 2 = leader_linetype_handle
        # 1 << 3 = leader_lineweight
        # 1 << 4 = landing_flag
        # 1 << 5 = landing_gap ???
        # 1 << 6 = dogleg_flag
        # 1 << 7 = dogleg_length
        # 1 << 8 = arrow_head_handle
        # 1 << 9 = arrow_head_size
        # 1 << 10 = content_type
        # 1 << 11 = text_style_handle
        # 1 << 12 = text_left_attachment_type (of MTEXT)
        # 1 << 13 = text_angle_type (of MTEXT)
        # 1 << 14 = text_alignment_type (of MTEXT)
        # 1 << 15 = text_color (of MTEXT)
        # 1 << 16 = ??? Text height (of MTEXT) ???
        # 1 << 17 = text_frame_flag
        # 1 << 18 = ??? Enable use of default MTEXT (from MLEADERSTYLE)
        # 1 << 19 = block_record_handle
        # 1 << 20 = block_color
        # 1 << 21 = block_scale_vector
        # 1 << 22 = block_rotation
        # 1 << 23 = block_connection_type
        # 1 << 24 = ??? Scale ???
        # 1 << 25 = text_right_attachment_type (of MTEXT)
        # 1 << 26 = ??? Text switch alignment type (of MTEXT) ???
        # 1 << 27 = text_attachment_direction (of MTEXT)
        # 1 << 28 = text_top_attachment_type (of MTEXT)
        # 1 << 29 = Text_bottom_attachment_type (of MTEXT)
        "leader_type": DXFAttr(170, default=1),
        "leader_line_color": DXFAttr(91, default=colors.BY_BLOCK_RAW_VALUE),
        "leader_linetype_handle": DXFAttr(341),
        "leader_lineweight": DXFAttr(171, default=const.LINEWEIGHT_BYBLOCK),
        "has_landing": DXFAttr(290, default=1),
        "has_dogleg": DXFAttr(291, default=1),
        "dogleg_length": DXFAttr(41, default=8),  # depend on $MEASUREMENT?
        # no handle is default arrow 'closed filled':
        "arrow_head_handle": DXFAttr(342),
        "arrow_head_size": DXFAttr(42, default=4),  # depend on $MEASUREMENT?
        "content_type": DXFAttr(172, default=2),
        # 0 = None
        # 1 = Block content
        # 2 = MTEXT content
        # 3 = TOLERANCE content
        # Text Content:
        "text_style_handle": DXFAttr(343),
        "text_left_attachment_type": DXFAttr(173, default=1),
        # Values 0-8 are used for the left/right attachment
        # point (attachment direction is horizontal), values 9-10 are used for the
        # top/bottom attachment points (attachment direction is vertical).
        # Attachment point is:
        # 0 = top of top text line,
        # 1 = middle of top text line,
        # 2 = middle of text,
        # 3 = middle of bottom text line,
        # 4 = bottom of bottom text line,
        # 5 = bottom text line,
        # 6 = bottom of top text line. Underline bottom line
        # 7 = bottom of top text line. Underline top line,
        # 8 = bottom of top text line. Underline all content,
        # 9 = center of text (y-coordinate only),
        # 10 = center of text (y-coordinate only), and overline top/underline
        # bottom content.
        "text_right_attachment_type": DXFAttr(95),  # like 173
        "text_angle_type": DXFAttr(174, default=1),
        # 0 = text angle is equal to last leader line segment angle
        # 1 = text is horizontal
        # 2 = text angle is equal to last leader line segment angle, but potentially
        #     rotated by 180 degrees so the right side is up for readability.
        "text_alignment_type": DXFAttr(175, default=2),
        "text_color": DXFAttr(92, default=colors.BY_BLOCK_RAW_VALUE),
        "has_frame_text": DXFAttr(292, default=0),
        # Block Content:
        "block_record_handle": DXFAttr(344),
        "block_color": DXFAttr(
            93, default=colors.BY_BLOCK_RAW_VALUE
        ),  # raw color
        "block_scale_vector": DXFAttr(
            10, xtype=XType.point3d, default=Vec3(1, 1, 1)
        ),
        "block_rotation": DXFAttr(43, default=0),  # in radians!!!
        "block_connection_type": DXFAttr(176, default=0),
        # 0 = center extents
        # 1 = insertion point
        "is_annotative": DXFAttr(293, default=0),
        # REPEAT "arrow_heads": DXF R2007+
        # arrow_head_index: 94, ???
        # arrow_head_handle: 345
        # END "arrow heads"
        # REPEAT "block attribs" (ATTDEF): DXF R2007+
        # attrib_handle: 330
        # attrib_index: 177, sequential index of the label in the collection
        # attrib_width: 44
        # attrib_text: 302, collision with group code (302, "LEADER{") in context data
        # END "block attribs"
        # Text Content:
        "is_text_direction_negative": DXFAttr(
            294, default=0, dxfversion=const.DXF2007
        ),
        "text_IPE_align": DXFAttr(178, default=0, dxfversion=const.DXF2007),
        "text_attachment_point": DXFAttr(
            179, default=1, dxfversion=const.DXF2007
        ),
        # 1 = left
        # 2 = center
        # 3 = right
        "scale": DXFAttr(45, default=1, dxfversion=const.DXF2007),
        "text_attachment_direction": DXFAttr(
            271, default=0, dxfversion=const.DXF2010
        ),
        # This defines whether the leaders attach to the left/right of the content
        # block/text, or attach to the top/bottom:
        # 0 = horizontal
        # 1 = vertical
        "text_bottom_attachment_direction": DXFAttr(
            272, default=9, dxfversion=const.DXF2010
        ),
        # like 173, but
        # 9 = center
        # 10= underline and center
        "text_top_attachment_direction": DXFAttr(
            273, default=9, dxfversion=const.DXF2010
        ),
        # like 173, but
        # 9 = center
        # 10= overline and center
        "leader_extend_to_text": DXFAttr(
            295, default=0, dxfversion=const.DXF2013
        ),
    },
)
acdb_mleader_group_codes = group_code_mapping(acdb_mleader)
CONTEXT_STR = "CONTEXT_DATA{"
LEADER_STR = "LEADER{"
LEADER_LINE_STR = "LEADER_LINE{"
START_CONTEXT_DATA = 300
END_CONTEXT_DATA = 301
START_LEADER = 302
END_LEADER = 303
START_LEADER_LINE = 304
END_LEADER_LINE = 305


def compile_context_tags(
    data: List["DXFTag"], stop_code: int
) -> List[Union["DXFTag", List]]:
    def build_structure(
        tag: "DXFTag", stop: int
    ) -> List[Union["DXFTag", List]]:
        collector = [tag]
        tag = next(tags)
        while tag.code != stop:
            if tag.code == START_LEADER:
                collector.append(build_structure(tag, END_LEADER))  # type: ignore
            # Group code 304 is used also for MTEXT content, therefore always
            # test for group code AND and value string:
            elif tag.code == START_LEADER_LINE and tag.value == LEADER_LINE_STR:
                collector.append(build_structure(tag, END_LEADER_LINE))  # type: ignore
            else:
                collector.append(tag)
            tag = next(tags)
        return collector  # type: ignore

    tags = iter(data)
    return build_structure(next(tags), stop_code)


ArrowHeadData = namedtuple("ArrowHeadData", "index, handle")
AttribData = namedtuple("AttribData", "handle, index, width, text")


@register_entity
class MultiLeader(DXFGraphic):
    DXFTYPE = "MULTILEADER"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_mleader)
    MIN_DXF_VERSION_FOR_EXPORT = const.DXF2000

    def __init__(self):
        super().__init__()
        self.context = MultiLeaderContext()
        self.arrow_heads: List[ArrowHeadData] = []
        self.block_attribs: List[AttribData] = []

    def _copy_data(self, entity: "DXFEntity") -> None:
        """Copy leaders"""
        assert isinstance(entity, MultiLeader)
        entity.context = copy.deepcopy(self.context)
        entity.arrow_heads = copy.deepcopy(self.arrow_heads)
        entity.block_attribs = copy.deepcopy(self.block_attribs)

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor is None:
            return dxf
        tags = processor.subclass_by_index(2)
        if tags:
            context = self.extract_context_data(tags)
            if context:
                try:
                    self.context = self.load_context(context)
                except const.DXFStructureError:
                    logger.info(
                        f"Context structure error in entity MULTILEADER(#{dxf.handle})"
                    )
        else:
            raise const.DXFStructureError(
                f"missing 'AcDbMLeader' subclass in MULTILEADER(#{dxf.handle})"
            )

        self.arrow_heads = self.extract_arrow_heads(tags)
        self.block_attribs = self.extract_block_attribs(tags)

        processor.fast_load_dxfattribs(
            dxf, acdb_mleader_group_codes, subclass=tags, recover=True
        )
        return dxf

    @staticmethod
    def extract_context_data(tags: Tags) -> List["DXFTag"]:
        start, end = None, None
        context_data = []
        for index, tag in enumerate(tags):
            if tag.code == START_CONTEXT_DATA:
                start = index
            elif tag.code == END_CONTEXT_DATA:
                end = index + 1

        if start and end:
            context_data = tags[start:end]
            # Remove context data!
            del tags[start:end]
        return context_data

    @staticmethod
    def load_context(data: List["DXFTag"]) -> "MultiLeaderContext":
        try:
            context = compile_context_tags(data, END_CONTEXT_DATA)
        except StopIteration:
            raise const.DXFStructureError
        else:
            return MultiLeaderContext.load(context)

    @staticmethod
    def extract_arrow_heads(data: Tags) -> List[ArrowHeadData]:
        def store_head():
            heads.append(
                ArrowHeadData(
                    collector.get(94, 0),  # arrow head index
                    collector.get(345, "0"),  # arrow head handle
                )
            )
            collector.clear()

        heads: List[ArrowHeadData] = []
        try:
            start = data.tag_index(94)
        except const.DXFValueError:
            return heads

        end = start
        collector = dict()
        for code, value in data.collect_consecutive_tags({94, 345}, start):
            end += 1
            collector[code] = value
            if code == 345:
                store_head()

        # Remove processed tags:
        del data[start:end]
        return heads

    @staticmethod
    def extract_block_attribs(data: Tags) -> List[AttribData]:
        def store_attrib():
            attribs.append(
                AttribData(
                    collector.get(330, "0"),  # ATTDEF handle
                    collector.get(177, 0),  # ATTDEF index
                    collector.get(44, 1.0),  # ATTDEF width
                    collector.get(302, ""),  # ATTDEF text (content)
                )
            )
            collector.clear()

        attribs: List[AttribData] = []
        try:
            start = data.tag_index(330)
        except const.DXFValueError:
            return attribs

        end = start
        collector: Dict[int, Any] = dict()
        for code, value in data.collect_consecutive_tags(
            {330, 177, 44, 302}, start
        ):
            end += 1
            if code == 330 and len(collector):
                store_attrib()
            collector[code] = value
        if len(collector):
            store_attrib()

        # Remove processed tags:
        del data[start:end]
        return attribs

    def preprocess_export(self, tagwriter: "TagWriter") -> bool:
        if self.context.is_valid:
            return True
        else:
            logger.debug(
                f"Ignore {str(self)} at DXF export, invalid context data."
            )
            return False

    def export_entity(self, tagwriter: "TagWriter") -> None:
        def write_handle_if_exist(code: int, name: str):
            handle = dxf.get(name)
            if handle is not None:
                write_tag2(code, handle)

        super().export_entity(tagwriter)
        dxf = self.dxf
        version = tagwriter.dxfversion
        write_tag2 = tagwriter.write_tag2

        write_tag2(100, acdb_mleader.name)
        write_tag2(270, dxf.version)
        self.context.export_dxf(tagwriter)

        # Export common MLEADER tags:
        # Don't use dxf.export_dxf_attribs() - all attributes should be written
        # even if equal to the default value:
        write_tag2(340, dxf.style_handle)
        write_tag2(90, dxf.property_override_flags)
        write_tag2(170, dxf.leader_type)
        write_tag2(91, dxf.leader_line_color)
        write_tag2(341, dxf.leader_linetype_handle)
        write_tag2(171, dxf.leader_lineweight)
        write_tag2(290, dxf.has_landing)
        write_tag2(291, dxf.has_dogleg)
        write_tag2(41, dxf.dogleg_length)
        # arrow_head_handle is None for default arrow 'closed filled':
        write_handle_if_exist(342, "arrow_head_handle")
        write_tag2(42, dxf.arrow_head_size)
        write_tag2(172, dxf.content_type)
        write_tag2(343, dxf.text_style_handle)  # mandatory!
        write_tag2(173, dxf.text_left_attachment_type)
        write_tag2(95, dxf.text_right_attachment_type)
        write_tag2(174, dxf.text_angle_type)
        write_tag2(175, dxf.text_alignment_type)
        write_tag2(92, dxf.text_color)
        write_tag2(292, dxf.has_frame_text)

        write_handle_if_exist(344, "block_record_handle")
        write_tag2(93, dxf.block_color)
        tagwriter.write_vertex(10, dxf.block_scale_vector)
        write_tag2(43, dxf.block_rotation)
        write_tag2(176, dxf.block_connection_type)
        write_tag2(293, dxf.is_annotative)
        if version >= const.DXF2007:
            self.export_arrow_heads(tagwriter)
            self.export_block_attribs(tagwriter)
            write_tag2(294, dxf.is_text_direction_negative)
            write_tag2(178, dxf.text_IPE_align)
            write_tag2(179, dxf.text_attachment_point)
            write_tag2(45, dxf.scale)

        if version >= const.DXF2010:
            write_tag2(271, dxf.text_attachment_direction)
            write_tag2(272, dxf.text_bottom_attachment_direction)
            write_tag2(273, dxf.text_top_attachment_direction)

        if version >= const.DXF2013:
            write_tag2(295, dxf.leader_extend_to_text)

    def export_arrow_heads(self, tagwriter: "TagWriter") -> None:
        for index, handle in self.arrow_heads:
            tagwriter.write_tag2(94, index)
            tagwriter.write_tag2(345, handle)

    def export_block_attribs(self, tagwriter: "TagWriter") -> None:
        for attrib in self.block_attribs:
            tagwriter.write_tag2(330, attrib.handle)
            tagwriter.write_tag2(177, attrib.index)
            tagwriter.write_tag2(44, attrib.width)
            tagwriter.write_tag2(302, attrib.text)

    def __virtual_entities__(self) -> Iterable["DXFGraphic"]:
        # As long as MLeader.virtual_entities() is not implemented,
        # use existing proxy graphic:
        if self.proxy_graphic:
            return ProxyGraphic(self.proxy_graphic, self.doc).virtual_entities()
        else:
            return []


class MultiLeaderContext:
    ATTRIBS = {
        40: "scale",
        10: "base_point",
        41: "text_height",
        140: "arrowhead_size",
        145: "landing_gap_size",
        174: "left_attachment",
        175: "right_attachment",
        176: "text_align_type",
        177: "attachment_type",
        110: "plane_origin",
        111: "plane_x_axis",
        112: "plane_y_axis",
        297: "plane_normal_reversed",
        272: "top_attachment",
        273: "bottom_attachment",
    }

    def __init__(self):
        self.leaders: List["Leader"] = []
        self.scale: float = 1.0  # overall scale
        self.base_point: Vec3 = NULLVEC
        self.text_height = 4.0
        self.arrowhead_size = 4.0
        self.landing_gap_size = 2.0
        self.left_attachment = 1
        self.right_attachment = 1
        self.text_align_type = 0  # 0=left, 1=center, 2=right
        self.attachment_type = 0  # 0=content extents, 1=insertion point
        self.mtext: Optional[MTextData] = None
        self.block: Optional[BlockData] = None
        self.plane_origin: Vec3 = NULLVEC
        self.plane_x_axis: Vec3 = X_AXIS
        self.plane_y_axis: Vec3 = Y_AXIS
        self.plane_normal_reversed: int = 0
        self.top_attachment = 9
        self.bottom_attachment = 9

    @classmethod
    def load(cls, context: List[Union["DXFTag", List]]) -> "MultiLeaderContext":
        assert context[0] == (START_CONTEXT_DATA, CONTEXT_STR)
        ctx = cls()
        content = None
        for tag in context:
            if isinstance(tag, list):  # Leader()
                ctx.leaders.append(Leader.load(tag))
                continue
            # parse context tags
            code, value = tag
            if content:
                if content.parse(code, value):
                    continue
                else:
                    content = None

            if code == 290 and value == 1:
                content = MTextData()
                ctx.mtext = content
            elif code == 296 and value == 1:
                content = BlockData()  # type: ignore
                ctx.block = content
            else:
                name = MultiLeaderContext.ATTRIBS.get(code)
                if name:
                    ctx.__setattr__(name, value)
        return ctx

    @property
    def is_valid(self) -> bool:
        return True

    def export_dxf(self, tagwriter: "TagWriter") -> None:
        write_tag2 = tagwriter.write_tag2
        write_vertex = tagwriter.write_vertex
        write_tag2(START_CONTEXT_DATA, CONTEXT_STR)
        # All MultiLeaderContext tags:
        write_tag2(40, self.scale)
        write_vertex(10, self.base_point)
        write_tag2(41, self.text_height)
        write_tag2(140, self.arrowhead_size)
        write_tag2(145, self.landing_gap_size)
        write_tag2(174, self.left_attachment)
        write_tag2(175, self.right_attachment)
        write_tag2(176, self.text_align_type)
        write_tag2(177, self.attachment_type)

        if self.mtext:
            write_tag2(290, 1)  # has mtext content
            self.mtext.export_dxf(tagwriter)
        else:
            write_tag2(290, 0)

        if self.block:
            write_tag2(296, 1)  # has block content
            self.block.export_dxf(tagwriter)
        else:
            write_tag2(296, 0)

        write_vertex(110, self.plane_origin)
        write_vertex(111, self.plane_x_axis)
        write_vertex(112, self.plane_y_axis)
        write_tag2(297, self.plane_normal_reversed)

        # Export Leader and LiederLine objects:
        for leader in self.leaders:
            leader.export_dxf(tagwriter)

        # Additional MultiLeaderContext tags:
        if tagwriter.dxfversion >= const.DXF2010:
            write_tag2(272, self.top_attachment)
            write_tag2(273, self.bottom_attachment)
        write_tag2(END_CONTEXT_DATA, "}")


class MTextData:
    ATTRIBS = {
        304: "default_content",
        11: "normal_direction",
        340: "style_handle",
        12: "location",
        13: "direction",
        42: "rotation",
        43: "boundary_width",
        44: "boundary_height",
        45: "line_space_factor",
        170: "line_space_style",
        90: "color",
        171: "alignment",
        172: "flow_direction",
        91: "bg_color",
        141: "bg_scale_factor",
        92: "bg_transparency",
        291: "has_bg_color",
        292: "has_bg_fill",
        173: "column_type",
        293: "use_auto_height",
        142: "column_width",
        143: "column_gutter_width",
        294: "column_flow_reversed",
        144: "column_sizes",  # multiple values
        295: "use_word_break",
    }

    def __init__(self):
        self.default_content: str = ""
        self.normal_direction: Vec3 = Z_AXIS
        self.style_handle = None  # handle of TextStyle() table entry
        self.location: Vec3 = NULLVEC
        self.direction: Vec3 = X_AXIS  # text direction
        self.rotation: float = 0.0  # in radians!
        self.boundary_width: float = 0.0
        self.boundary_height: float = 0.0
        self.line_space_factor: float = 1.0
        self.line_space_style: int = 1  # 1=at least, 2=exactly
        self.color: int = colors.BY_BLOCK_RAW_VALUE
        self.alignment: int = 1  # 1=left, 2=center, 3=right
        self.flow_direction: int = 1  # 1=horiz, 3=vert, 6=by style
        self.bg_color: int = -939524096  # use window background color? (CMC)
        self.bg_scale_factor: float = 1.5
        self.bg_transparency: int = 0
        self.has_bg_color: int = 0
        self.has_bg_fill: int = 0
        self.column_type: int = 0  # unknown values
        self.use_auto_height: int = 0
        self.column_width: float = 0.0
        self.column_gutter_width: float = 0.0
        self.column_flow_reversed: int = 0
        self.column_sizes: List[float] = []  # heights?
        self.use_word_break: int = 1

    def parse(self, code: int, value) -> bool:
        # return True if data belongs to mtext else False (end of mtext section)
        if code == 144:
            self.column_sizes.append(value)
            return True
        attrib = MTextData.ATTRIBS.get(code)
        if attrib:
            self.__setattr__(attrib, value)
        return bool(attrib)

    def export_dxf(self, tagwriter: "TagWriter") -> None:
        write_tag2 = tagwriter.write_tag2
        write_vertex = tagwriter.write_vertex
        write_tag2(304, self.default_content)
        write_vertex(11, self.normal_direction)
        if self.style_handle:
            write_tag2(340, self.style_handle)
        else:
            # Do not write None, but "0" is also not valid!
            # DXF structure error should be detected before export.
            write_tag2(340, "0")
        write_vertex(12, self.location)
        write_vertex(13, self.direction)
        write_tag2(42, self.rotation)
        write_tag2(43, self.boundary_width)
        write_tag2(44, self.boundary_height)
        write_tag2(45, self.line_space_factor)
        write_tag2(170, self.line_space_style)
        write_tag2(90, self.color)
        write_tag2(171, self.alignment)
        write_tag2(172, self.flow_direction)
        write_tag2(91, self.bg_color)
        write_tag2(141, self.bg_scale_factor)
        write_tag2(92, self.bg_transparency)
        write_tag2(291, self.has_bg_color)
        write_tag2(292, self.has_bg_fill)
        write_tag2(173, self.column_type)
        write_tag2(293, self.use_auto_height)
        write_tag2(142, self.column_width)
        write_tag2(143, self.column_gutter_width)
        write_tag2(294, self.column_flow_reversed)
        for size in self.column_sizes:
            write_tag2(144, size)
        write_tag2(295, self.use_word_break)


class BlockData:
    ATTRIBS = {
        341: "block_record_handle",
        14: "normal_direction",
        15: "location",
        16: "scale",
        46: "rotation",
        93: "color",
    }

    def __init__(self):
        self.block_record_handle = None
        self.normal_direction: Vec3 = Z_AXIS
        self.location: Vec3 = NULLVEC
        self.scale: Vec3 = Vec3(1, 1, 1)
        self.rotation: float = 0  # in radians!
        self.color: int = colors.BY_BLOCK_RAW_VALUE
        # The transformation matrix is stored in transposed order
        # of ezdxf.math.Matrix44()!
        self._matrix: List[float] = []  # group code 47 x 16

    @property
    def matrix44(self) -> Matrix44:
        m = Matrix44(self._matrix)
        m.transpose()
        return m

    @matrix44.setter
    def matrix44(self, m: Matrix44) -> None:
        m = m.copy()
        m.transpose()
        self._matrix = list(m)

    def parse(self, code: int, value) -> bool:
        attrib = BlockData.ATTRIBS.get(code)
        if attrib:
            self.__setattr__(attrib, value)
        elif code == 47:
            self._matrix.append(value)
        else:
            return False
        # return True if data belongs to block else False (end of block section)
        return True

    def export_dxf(self, tagwriter: "TagWriter") -> None:
        write_tag2 = tagwriter.write_tag2
        write_vertex = tagwriter.write_vertex
        if self.block_record_handle:
            write_tag2(341, self.block_record_handle)
        else:
            # Do not write None, but "0" is also not valid!
            # DXF structure error should be detected before export.
            write_tag2(341, "0")
        write_vertex(14, self.normal_direction)
        write_vertex(15, self.location)
        write_vertex(16, self.scale)
        write_tag2(46, self.rotation)
        write_tag2(93, self.color)
        for value in self._matrix:
            write_tag2(47, value)


class Leader:
    def __init__(self):
        self.lines: List["LeaderLine"] = []
        self.has_last_leader_line: int = 0  # group code 290
        self.has_dogleg_vector: int = 0  # group code 291
        self.last_leader_point: Vec3 = NULLVEC  # group code (10, 20, 30)
        self.dogleg_vector: Vec3 = X_AXIS  # group code (11, 21, 31)
        self.dogleg_length: float = 1.0  # group code 40
        self.index: int = 0  # group code 90
        self.attachment_direction: int = 0  # group code 271, R21010+
        self.breaks = []  # group code 12, 13 - multiple breaks possible!

    @classmethod
    def load(cls, context: List[Union["DXFTag", List]]):
        assert context[0] == (START_LEADER, LEADER_STR)
        leader = cls()
        for tag in context:
            if isinstance(tag, list):  # LeaderLine()
                leader.lines.append(LeaderLine.load(tag))
                continue

            code, value = tag
            if code == 290:
                leader.has_last_leader_line = value
            elif code == 291:
                leader.has_dogleg_vector = value
            elif code == 10:
                leader.last_leader_point = value
            elif code == 11:
                leader.dogleg_vector = value
            elif code == 40:
                leader.dogleg_length = value
            elif code == 90:
                leader.index = value
            elif code == 271:
                leader.attachment_direction = value
            elif code in (12, 13):
                leader.breaks.append(value)

        return leader

    def export_dxf(self, tagwriter: "TagWriter") -> None:
        write_tag2 = tagwriter.write_tag2
        write_vertex = tagwriter.write_vertex

        write_tag2(START_LEADER, LEADER_STR)
        write_tag2(290, self.has_last_leader_line)
        write_tag2(291, self.has_dogleg_vector)
        if self.has_last_leader_line:
            write_vertex(10, self.last_leader_point)
        if self.has_dogleg_vector:
            write_vertex(11, self.dogleg_vector)

        code = 0
        for vertex in self.breaks:
            # write alternate group code 12 and 13
            write_vertex(12 + code, vertex)
            code = 1 - code
        write_tag2(90, self.index)
        write_tag2(40, self.dogleg_length)

        # Export leader lines:
        for line in self.lines:
            line.export_dxf(tagwriter)

        if tagwriter.dxfversion >= const.DXF2010:
            write_tag2(271, self.attachment_direction)
        write_tag2(END_LEADER, "}")


class LeaderLine:
    def __init__(self):
        self.vertices: List[Vec3] = []
        self.breaks: Optional[List[Union[int, Vec3]]] = None
        # Breaks: 90, 11, 12, [11, 12, ...] [, 90, 11, 12 [11, 12, ...]]
        # group code 90 = break index
        # group code 11 = start vertex of break
        # group code 12 = end vertex of break
        # multiple breaks per index possible
        self.index: int = 0  # group code 91
        self.color: int = colors.BY_BLOCK_RAW_VALUE  # group code 92
        # R2010+: override properties see ODA DWG pg. 214-215

    @classmethod
    def load(cls, tags: List["DXFTag"]):
        assert tags[0] == (START_LEADER_LINE, LEADER_LINE_STR)
        line = LeaderLine()
        vertices = line.vertices
        breaks = []
        for code, value in tags:
            if code == 10:
                vertices.append(value)
            elif code in (90, 11, 12):
                breaks.append(value)
            elif code == 91:
                line.index = value
            elif code == 92:
                line.color = value
        if breaks:
            line.breaks = breaks
        return line

    def export_dxf(self, tagwriter: "TagWriter") -> None:
        write_tag2 = tagwriter.write_tag2
        write_vertex = tagwriter.write_vertex

        write_tag2(START_LEADER_LINE, LEADER_LINE_STR)
        for vertex in self.vertices:
            write_vertex(10, vertex)
        if self.breaks:
            code = 0
            for value in self.breaks:
                if isinstance(value, int):
                    # break index
                    write_tag2(90, value)
                else:
                    # 11 .. start vertex of break
                    # 12 .. end vertex of break
                    write_vertex(11 + code, value)
                    code = 1 - code
        write_tag2(91, self.index)
        write_tag2(92, self.color)
        write_tag2(END_LEADER_LINE, "}")


@register_entity
class MLeader(MultiLeader):
    DXFTYPE = "MLEADER"


acdb_mleader_style = DefSubclass(
    "AcDbMLeaderStyle",
    {
        "unknown1": DXFAttr(179, default=2),
        "content_type": DXFAttr(170, default=2),
        "draw_mleader_order_type": DXFAttr(171, default=1),
        "draw_leader_order_type": DXFAttr(172, default=0),
        "max_leader_segments_points": DXFAttr(90, default=2),
        "first_segment_angle_constraint": DXFAttr(40, default=0.0),
        "second_segment_angle_constraint": DXFAttr(41, default=0.0),
        "leader_type": DXFAttr(173, default=1),
        "leader_line_color": DXFAttr(91, default=-1056964608),
        # raw color: BY_BLOCK
        # raw color: BY_BLOCK
        "leader_linetype_handle": DXFAttr(340),
        "leader_lineweight": DXFAttr(92, default=-2),
        "has_landing": DXFAttr(290, default=1),
        "landing_gap": DXFAttr(42, default=2.0),
        "has_dogleg": DXFAttr(291, default=1),
        "dogleg_length": DXFAttr(43, default=8),
        "name": DXFAttr(3, default="Standard"),
        # no handle is default arrow 'closed filled':
        "arrow_head_handle": DXFAttr(341),
        "arrow_head_size": DXFAttr(44, default=4),
        "default_text_content": DXFAttr(300, default=""),
        "text_style_handle": DXFAttr(342),
        "text_left_attachment_type": DXFAttr(174, default=1),
        "text_angle_type": DXFAttr(175, default=1),
        "text_alignment_type": DXFAttr(176, default=0),
        "text_right_attachment_type": DXFAttr(178, default=1),
        "text_color": DXFAttr(93, default=-1056964608),  # raw color: BY_BLOCK
        "text_height": DXFAttr(45, default=4),
        "has_frame_text": DXFAttr(292, default=0),
        "text_align_always_left": DXFAttr(297, default=0),
        "align_space": DXFAttr(46, default=4),
        "has_block_scaling": DXFAttr(293),
        "block_record_handle": DXFAttr(343),
        "block_color": DXFAttr(94, default=-1056964608),  # raw color: BY_BLOCK
        "block_scale_x": DXFAttr(47, default=1),
        "block_scale_y": DXFAttr(49, default=1),
        "block_scale_z": DXFAttr(140, default=1),
        "has_block_rotation": DXFAttr(294, default=1),
        "block_rotation": DXFAttr(141, default=0),
        "block_connection_type": DXFAttr(177, default=0),
        "scale": DXFAttr(142, default=1),
        "overwrite_property_value": DXFAttr(295, default=0),
        "is_annotative": DXFAttr(296, default=0),
        "break_gap_size": DXFAttr(143, default=3.75),
        # 0 = Horizontal; 1 = Vertical:
        "text_attachment_direction": DXFAttr(271, default=0),
        # 9 = Center; 10 = Underline and Center:
        "text_bottom__attachment_direction": DXFAttr(272, default=9),
        # 9 = Center; 10 = Overline and Center:
        "text_top_attachment_direction": DXFAttr(273, default=9),
        "unknown2": DXFAttr(298, optional=True),  # boolean flag ?
    },
)
acdb_mleader_style_group_codes = group_code_mapping(acdb_mleader_style)


@register_entity
class MLeaderStyle(DXFObject):
    DXFTYPE = "MLEADERSTYLE"
    DXFATTRIBS = DXFAttributes(base_class, acdb_mleader_style)
    MIN_DXF_VERSION_FOR_EXPORT = const.DXF2000

    def load_dxf_attribs(
        self, processor: SubclassProcessor = None
    ) -> "DXFNamespace":
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_mleader_style_group_codes, subclass=1
            )
        return dxf

    def export_entity(self, tagwriter: "TagWriter") -> None:
        super().export_entity(tagwriter)
        tagwriter.write_tag2(const.SUBCLASS_MARKER, acdb_mleader_style.name)
        self.dxf.export_dxf_attribs(
            tagwriter, acdb_mleader_style.attribs.keys()
        )


class MLeaderStyleCollection(ObjectCollection):
    def __init__(self, doc: "Drawing"):
        super().__init__(
            doc, dict_name="ACAD_MLEADERSTYLE", object_type="MLEADERSTYLE"
        )
        self.create_required_entries()

    def create_required_entries(self) -> None:
        for name in ("Standard",):
            if name not in self.object_dict:
                mleader_style = self.new(name)
                # set standard text style
                text_style = self.doc.styles.get("Standard")
                mleader_style.dxf.text_style_handle = text_style.dxf.handle



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
from typing import Optional

# Documentation: https://ezdxf.mozman.at/docs/concepts/units.html#insunits

MSP_METRIC_UNITS_FACTORS = {
    'km': .001,
    'm': 1.0,
    'dm': 10.,
    'cm': 100.,
    'mm': 1000.,
    'µm': 1000000.,
    'yd': 1.093613298,
    'ft': 3.280839895,
    'in': 39.37007874,
    'mi': 0.00062137119,
}

IN = 1
FT = 2
MI = 3
MM = 4
CM = 5
M = 6
KM = 7
YD = 10
DM = 14

IMPERIAL_UNITS = {1, 2, 3, 8, 9, 10, 21, 22, 23, 24}

# Conversion factor from meters to unit
# 1 meter is ... [unit]
METER_FACTOR = [
    None,  # 0 = Unitless - not supported
    39.37007874,  # 1 = Inches
    3.280839895,  # 2 = Feet
    0.00062137119,  # 3 = Miles
    1000.0,  # 4 = Millimeters
    100.0,  # 5 = Centimeters
    1.0,  # 6 = Meters
    0.001,  # 7 = Kilometers
    None,  # 8 = Microinches = 1e-6 in
    None,  # 9 = Mils = 0.001 in
    1.093613298,  # 10 = Yards
    10000000000.0,  # 11 = Angstroms = 1e-10m
    1000000000.0,  # 12 = Nanometers = 1e-9m
    1000000.0,  # 13 = Microns = 1e-6m
    10.0,  # 14 = Decimeters = 0.1m
    0.1,  # 15 = Decameters = 10m
    0.01,  # 16 = Hectometers = 100m
    0.000000001,  # 17 = Gigameters = 1e+9 m
    1.0 / 149597870700,  # 18 = Astronomical units = 149597870700m
    1.0 / 9.46e15,  # 19 = Light years = 9.46e15 m
    1.0 / 3.09e16,  # 20 = Parsecs =  3.09e16 m
    None,  # 21 = US Survey Feet
    None,  # 22 = US Survey Inch
    None,  # 23 = US Survey Yard
    None,  # 24 = US Survey Mile
]


class DrawingUnits:
    def __init__(self, base: float = 1., unit: str = 'm'):
        self.base = float(base)
        self.unit = unit.lower()
        self._units = MSP_METRIC_UNITS_FACTORS
        self._msp_factor = base * self._units[self.unit]

    def factor(self, unit: str = 'm') -> float:
        return self._msp_factor / self._units[unit.lower()]

    def __call__(self, unit: str) -> float:
        return self.factor(unit)


class PaperSpaceUnits:
    def __init__(self, msp=DrawingUnits(), unit: str = 'mm', scale: float = 1):
        self.unit = unit.lower()
        self.scale = scale
        self._msp = msp
        self._psp = DrawingUnits(1, self.unit)

    def from_msp(self, value: float, unit: str):
        drawing_units = value * self._msp(unit.lower())
        return drawing_units / (self._msp(self.unit) * self.scale)

    def to_msp(self, value: float, unit: str):
        paper_space_units = value * self.scale * self._psp.factor(unit)
        model_space_units = paper_space_units * self._msp.factor(self.unit)
        return model_space_units


# Layout units are stored as enum in the associated BLOCK_RECORD: BlockRecord.dxf.units
# or as  optional XDATA for all DXF versions
# 1000: "ACAD"
# 1001: "DesignCenter Data" (optional)
# 1002: "{"
# 1070: Autodesk Design Center version number
# 1070: Insert units: like 'units'
# 1002: "}"

# The units of the modelspace block record is always 0, the real modelspace
# units and therefore the document units are stored as enum in the header var
# $INSUNITS

# units stored as enum in BlockRecord.dxf.units
# 0 = Unitless
# 1 = Inches
# 2 = Feet
# 3 = Miles
# 4 = Millimeters
# 5 = Centimeters
# 6 = Meters
# 7 = Kilometers
# 8 = Microinches = 1e-6 in
# 9 = Mils = 0.001 in
# 10 = Yards
# 11 = Angstroms = 1e-10m
# 12 = Nanometers = 1e-9m
# 13 = Microns = 1e-6m
# 14 = Decimeters = 0.1m
# 15 = Decameters = 10m
# 16 = Hectometers = 100m
# 17 = Gigameters = 1e+9 m
# 18 = Astronomical units = 149597870700m = 1.58125074e−5 ly =  4.84813681e−6 Parsec
# 19 = Light years = 9.46e15 m
# 20 = Parsecs =  3.09e16 m
# 21 = US Survey Feet
# 22 = US Survey Inch
# 23 = US Survey Yard
# 24 = US Survey Mile
_unit_spec = [
    None, 'in', 'ft', 'mi', 'mm', 'cm', 'm', 'km',
    'µin', 'mil', 'yd', 'Å', 'nm', 'µm', 'dm', 'dam', 'hm', 'gm',
    'au', 'ly', 'pc',
    None, None, None, None,
]


def decode(enum: int) -> Optional[str]:
    return _unit_spec[int(enum)]


def conversion_factor(source_units: int, target_units: int) -> float:
    """ Returns the conversion factor to represent `source_units` in
    `target_units`.

    E.g. millimeter in centimeter :code:`conversion_factor(MM, CM)` returns 0.1,
    because 1 mm = 0.1 cm

    """
    try:
        source_factor = METER_FACTOR[source_units]
        target_factor = METER_FACTOR[target_units]
        if source_factor is None or target_factor is None:
            raise TypeError('Unsupported conversion.')
        return target_factor / source_factor

    except IndexError:
        raise ValueError('Invalid unit enum.')



# Copyright (c) 2011-2021, Manfred Moitzi
# License: MIT License
from typing import TYPE_CHECKING, Iterable, Optional, List, Iterator, cast
from itertools import chain
import logging
from .types import tuples_to_tags, NONE_TAG
from .tags import Tags, DXFTag
from .const import DXFStructureError, DXFValueError, DXFKeyError
from .types import (
    APP_DATA_MARKER,
    SUBCLASS_MARKER,
    XDATA_MARKER,
    EMBEDDED_OBJ_MARKER,
    EMBEDDED_OBJ_STR,
)
from .types import is_app_data_marker, is_embedded_object_marker
from .tagger import internal_tag_compiler

logger = logging.getLogger("ezdxf")

if TYPE_CHECKING:
    from ezdxf.eztypes import IterableTags


class ExtendedTags:
    """Manages DXF tags located in sub structures:

        - Subclasses
        - AppData
        - Extended Data (XDATA)
        - Embedded objects

    Args:
        tags: iterable of type DXFTag()
        legacy: flag for DXF R12 tags

    """

    __slots__ = ("subclasses", "appdata", "xdata", "embedded_objects")

    def __init__(self, tags: Iterable[DXFTag] = None, legacy=False):
        if isinstance(tags, str):
            raise DXFValueError(
                "use ExtendedTags.from_text() to create tags from a string."
            )

        # code == 102, keys are "{<arbitrary name>", values are Tags()
        self.appdata: List[Tags] = list()

        # code == 100, keys are "subclass-name", values are Tags()
        self.subclasses: List[Tags] = list()

        # code >= 1000, keys are "APPID", values are Tags()
        self.xdata: List[Tags] = list()

        # Store embedded objects as list, but embedded objects are rare, so
        # storing an empty list for every DXF entity is waste of memory.
        # Support for multiple embedded objects is maybe future proof, but
        # for now only one embedded object per entity is used.
        self.embedded_objects: Optional[List[Tags]] = None

        if tags is not None:
            self._setup(iter(tags))
            if legacy:
                self.legacy_repair()

    def legacy_repair(self):
        """Legacy (DXF R12) tags handling and repair."""
        self.flatten_subclasses()
        # ... and we can do some checks:
        # DXF R12 does not support (102, '{APPID') ... structures
        if len(self.appdata):
            # Just a debug message, do not delete appdata, this would corrupt
            # the data structure.
            self.debug("Found application defined entity data in DXF R12.")

        # That is really unlikely, but...
        if self.embedded_objects is not None:
            # Removing embedded objects from DXF R12 does not corrupt the
            # data structure:
            self.embedded_objects = None
            self.debug("Found embedded object in DXF R12.")

    def flatten_subclasses(self):
        """Flatten subclasses in legacy mode (DXF R12).

        There exists DXF R12 with subclass markers, technical incorrect but
        works if the reader ignore subclass marker tags, unfortunately ezdxf
        tries to use this subclass markers and therefore R12 parsing by ezdxf
        does not work without removing these subclass markers.

        This method removes all subclass markers and flattens all subclasses
        into ExtendedTags.noclass.

        """
        if len(self.subclasses) < 2:
            return
        noclass = self.noclass
        for subclass in self.subclasses[1:]:
            # Exclude first tag (100, subclass marker):
            noclass.extend(subclass[1:])
        self.subclasses = [noclass]
        self.debug("Removed subclass marker from entity for DXF R12.")

    def debug(self, msg: str) -> None:
        msg += f" <{self.entity_name()}>"
        logger.debug(msg)

    def entity_name(self) -> str:
        try:
            handle = f"(#{self.get_handle()})"
        except DXFValueError:
            handle = ""
        return self.dxftype() + handle

    def __copy__(self) -> "ExtendedTags":
        """Shallow copy."""

        def copy(tag_lists):
            return [tags.clone() for tags in tag_lists]

        clone = self.__class__()
        clone.appdata = copy(self.appdata)
        clone.subclasses = copy(self.subclasses)
        clone.xdata = copy(self.xdata)
        if self.embedded_objects is not None:
            clone.embedded_objects = copy(self.embedded_objects)
        return clone

    clone = __copy__

    def __getitem__(self, index) -> Tags:
        return self.noclass[index]

    @property
    def noclass(self) -> Tags:
        """Short cut to access first subclass."""
        return self.subclasses[0]

    def get_handle(self) -> str:
        """Returns handle as hex string."""
        return self.noclass.get_handle()

    def dxftype(self) -> str:
        """Returns DXF type as string like "LINE"."""
        return self.noclass[0].value

    def replace_handle(self, handle: str) -> None:
        """Replace the existing entity handle by a new value."""
        self.noclass.replace_handle(handle)

    def _setup(self, tags: Iterator[DXFTag]) -> None:
        def is_end_of_class(tag):
            # fast path
            if tag.code not in {
                SUBCLASS_MARKER,
                EMBEDDED_OBJ_MARKER,
                XDATA_MARKER,
            }:
                return False
            else:
                # really an embedded object
                if (
                    tag.code == EMBEDDED_OBJ_MARKER
                    and tag.value != EMBEDDED_OBJ_STR
                ):
                    return False
                else:
                    return True

        def collect_base_class() -> DXFTag:
            """The base class contains AppData, but not XDATA and ends with
            SUBCLASS_MARKER, XDATA_MARKER or EMBEDDED_OBJ_MARKER.
            """
            # All subclasses begin with (100, subclass name) EXCEPT DIMASSOC
            # has a subclass starting with: (1, AcDbOsnapPointRef)
            # This special subclass is ignored by ezdxf, content is included in
            # the preceding subclass: (100, AcDbDimAssoc)
            # TEXT contains 2x the (100, AcDbText).
            #
            # Therefore it is not possible to use an (ordered) dict with
            # the subclass name as key, but usual use case is access by
            # numerical index.

            data = Tags()
            try:
                while True:
                    tag = next(tags)
                    if is_app_data_marker(tag):
                        app_data_pos = len(self.appdata)
                        data.append(DXFTag(tag.code, app_data_pos))
                        collect_app_data(tag)
                    elif is_end_of_class(tag):
                        self.subclasses.append(data)
                        return tag
                    else:
                        data.append(tag)
            except StopIteration:
                pass
            self.subclasses.append(data)
            return NONE_TAG

        def collect_subclass(starttag: DXFTag) -> DXFTag:
            """A subclass does NOT contain AppData or XDATA, and ends with
            SUBCLASS_MARKER, XDATA_MARKER or EMBEDDED_OBJ_MARKER.
            """
            # All subclasses begin with (100, subclass name)
            # for exceptions and rant see: collect_base_class()

            data = Tags([starttag])
            try:
                while True:
                    tag = next(tags)
                    # removed app data collection in subclasses
                    if is_end_of_class(tag):
                        self.subclasses.append(data)
                        return tag
                    else:
                        data.append(tag)
            except StopIteration:
                pass
            self.subclasses.append(data)
            return NONE_TAG

        def collect_app_data(starttag: DXFTag) -> None:
            """AppData can't contain XDATA or subclasses.

            AppData can only appear in the first unnamed subclass
            """
            data = Tags([starttag])
            # Alternative closing tag 'APPID}':
            closing_strings = ("}", starttag.value[1:] + "}")  # type: ignore
            while True:
                try:
                    tag = next(tags)
                except StopIteration:
                    raise DXFStructureError(
                        "Missing closing (102, '}') tag in appdata structure."
                    )
                data.append(tag)
                if (tag.code == APP_DATA_MARKER) and (
                    tag.value in closing_strings
                ):
                    break
                    # Other (102, ) tags are treated as usual DXF tags.
            self.appdata.append(data)

        def collect_xdata(starttag: DXFTag) -> DXFTag:
            """XDATA is always at the end of the entity even if an embedded
            object is present and can not contain AppData or subclasses.

            """
            data = Tags([starttag])
            try:
                while True:
                    tag = next(tags)
                    if tag.code == XDATA_MARKER:
                        self.xdata.append(data)
                        return tag
                    else:
                        data.append(tag)
            except StopIteration:
                pass
            self.xdata.append(data)
            return NONE_TAG

        def collect_embedded_object(starttag: DXFTag) -> DXFTag:
            """Since AutoCAD 2018, DXF entities can contain embedded objects
            starting with a (101, 'Embedded Object') tag.

            All embedded object data is collected in a simple Tags() object,
            no subclass app data or XDATA processing is done.
            ezdxf does not use or modify the embedded object data, the data is
            just stored and written out as it is.

            self.embedded_objects = [
                1. embedded object as Tags(),
                2. embedded object as Tags(),
                ...
            ]

            Support for multiple embedded objects is maybe future proof, but
            for now only one embedded object per entity is used.

            """
            if self.embedded_objects is None:
                self.embedded_objects = list()
            data = Tags([starttag])
            try:
                while True:
                    tag = next(tags)
                    if (
                        is_embedded_object_marker(tag)
                        or tag.code == XDATA_MARKER
                    ):
                        # Another embedded object found: maybe in the future
                        # DXF entities can contain more than one embedded
                        # object.
                        self.embedded_objects.append(data)
                        return tag
                    else:
                        data.append(tag)
            except StopIteration:
                pass
            self.embedded_objects.append(data)
            return NONE_TAG

        # Preceding tags without a subclass
        tag = collect_base_class()
        while tag.code == SUBCLASS_MARKER:
            tag = collect_subclass(tag)

        while is_embedded_object_marker(tag):
            tag = collect_embedded_object(tag)

        # XDATA appear after an embedded object
        while tag.code == XDATA_MARKER:
            tag = collect_xdata(tag)

        if tag is not NONE_TAG:
            raise DXFStructureError(
                "Unexpected tag '%r' at end of entity." % tag
            )

    def __iter__(self) -> Iterator[DXFTag]:
        for subclass in self.subclasses:
            for tag in subclass:
                if tag.code == APP_DATA_MARKER and isinstance(tag.value, int):
                    yield from self.appdata[tag.value]
                else:
                    yield tag
        yield from chain.from_iterable(self.xdata)
        if self.embedded_objects is not None:
            yield from chain.from_iterable(self.embedded_objects)

    def get_subclass(self, name: str, pos: int = 0) -> Tags:
        """Get subclass `name`.

        Args:
            name: subclass name as string like "AcDbEntity"
            pos: start searching at subclass `pos`.

        """
        for index, subclass in enumerate(self.subclasses):
            try:
                if (index >= pos) and (subclass[0].value == name):
                    return subclass
            except IndexError:
                pass  # subclass[0]: ignore empty subclasses

        raise DXFKeyError(f'Subclass "{name}" does not exist.')

    def has_subclass(self, name: str) -> bool:
        for subclass in self.subclasses:
            try:
                if subclass[0].value == name:
                    return True
            except IndexError:
                pass  # ignore empty subclasses
        return False

    def has_xdata(self, appid: str) -> bool:
        """``True`` if has XDATA for `appid`."""
        return any(xdata[0].value == appid for xdata in self.xdata)

    def get_xdata(self, appid: str) -> Tags:
        """Returns XDATA for `appid` as :class:`Tags`."""
        for xdata in self.xdata:
            if xdata[0].value == appid:
                return xdata
        raise DXFValueError(f'No extended data for APPID "{appid}".')

    def set_xdata(self, appid: str, tags: "IterableTags") -> None:
        """Set `tags` as XDATA for `appid`."""
        xdata = self.get_xdata(appid)
        xdata[1:] = tuples_to_tags(tags)

    def new_xdata(self, appid: str, tags: "IterableTags" = None) -> Tags:
        """Append a new XDATA block.

        Assumes that no XDATA block with the same `appid` already exist::

            try:
                xdata = tags.get_xdata('EZDXF')
            except ValueError:
                xdata = tags.new_xdata('EZDXF')
        """
        xtags = Tags([DXFTag(XDATA_MARKER, appid)])
        if tags is not None:
            xtags.extend(tuples_to_tags(tags))
        self.xdata.append(xtags)
        return xtags

    def has_app_data(self, appid: str) -> bool:
        """``True`` if has application defined data for `appid`."""
        return any(appdata[0].value == appid for appdata in self.appdata)

    def get_app_data(self, appid: str) -> Tags:
        """Returns application defined data for `appid` as :class:`Tags`
        including marker tags."""
        for appdata in self.appdata:
            if appdata[0].value == appid:
                return appdata
        raise DXFValueError(
            f'Application defined group "{appid}" does not exist.'
        )

    def get_app_data_content(self, appid: str) -> Tags:
        """Returns application defined data for `appid` as :class:`Tags`
        without first and last marker tag.
        """
        return Tags(self.get_app_data(appid)[1:-1])

    def set_app_data_content(self, appid: str, tags: "IterableTags") -> None:
        """Set application defined data for `appid` for already exiting data."""
        app_data = self.get_app_data(appid)
        app_data[1:-1] = tuples_to_tags(tags)

    def new_app_data(
        self, appid: str, tags: "IterableTags" = None, subclass_name: str = None
    ) -> Tags:
        """Append a new application defined data to subclass `subclass_name`.

        Assumes that no app data block with the same `appid` already exist::

            try:
                app_data = tags.get_app_data('{ACAD_REACTORS', tags)
            except ValueError:
                app_data = tags.new_app_data('{ACAD_REACTORS', tags)

        """
        if not appid.startswith("{"):
            raise DXFValueError("Appid has to start with '{'.")

        app_tags = Tags(
            [
                DXFTag(APP_DATA_MARKER, appid),
                DXFTag(APP_DATA_MARKER, "}"),
            ]
        )
        if tags is not None:
            app_tags[1:1] = tuples_to_tags(tags)

        if subclass_name is None:
            subclass = self.noclass
        else:
            # raises KeyError, if not exist
            subclass = self.get_subclass(subclass_name, 1)
        app_data_pos = len(self.appdata)
        subclass.append(DXFTag(APP_DATA_MARKER, app_data_pos))
        self.appdata.append(app_tags)
        return app_tags

    @classmethod
    def from_text(cls, text: str, legacy=False) -> "ExtendedTags":
        """Create :class:`ExtendedTags` from DXF text."""
        return cls(internal_tag_compiler(text), legacy=legacy)



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
from typing import Iterable, Optional, cast, TYPE_CHECKING, List
import abc
import math
from ezdxf.entities import DXFEntity, Insert
from ezdxf.lldxf import const
from ezdxf.math import Vec3, UCS, Z_AXIS, X_AXIS
from ezdxf.path import Path, make_path, from_vertices
from ezdxf.render import MeshBuilder, MeshVertexMerger, TraceBuilder
from ezdxf.protocols import SupportsVirtualEntities, virtual_entities

from ezdxf.tools.text import (
    TextLine,
    unified_alignment,
    plain_text,
    text_wrap,
)
from ezdxf.tools import fonts

if TYPE_CHECKING:
    from ezdxf.eztypes import LWPolyline, Polyline, MText

__all__ = [
    "make_primitive",
    "recursive_decompose",
    "to_primitives",
    "to_vertices",
    "to_control_vertices",
    "to_paths",
    "to_meshes",
]


class Primitive(abc.ABC):
    """It is not efficient to create the Path() or MeshBuilder() representation
    by default. For some entities the it's just not needed (LINE, POINT) and for
    others the builtin flattening() method is more efficient or accurate than
    using a Path() proxy object. (ARC, CIRCLE, ELLIPSE, SPLINE).

    The `max_flattening_distance` defines the max distance between the
    approximation line and the original curve. Use argument
    `max_flattening_distance` to override the default value, or set the value
    by direct attribute access.

    """

    max_flattening_distance: float = 0.01

    def __init__(self, entity: DXFEntity, max_flattening_distance: Optional[float] = None):
        self.entity: DXFEntity = entity
        # Path representation for linear entities:
        self._path: Optional[Path] = None
        # MeshBuilder representation for mesh based entities:
        # PolygonMesh, PolyFaceMesh, Mesh
        self._mesh: Optional[MeshBuilder] = None
        if max_flattening_distance is not None:
            self.max_flattening_distance = max_flattening_distance

    @property
    def is_empty(self) -> bool:
        """Returns `True` if represents an empty primitive which do not
        yield any vertices.

        """
        if self._mesh:
            return len(self._mesh.vertices) == 0
        return self.path is None  # on demand calculations!

    @property
    def path(self) -> Optional[Path]:
        """:class:`~ezdxf.path.Path` representation or ``None``,
        idiom to check if is a path representation (could be empty)::

            if primitive.path is not None:
                process(primitive.path)

        """
        return None

    @property
    def mesh(self) -> Optional[MeshBuilder]:
        """:class:`~ezdxf.render.mesh.MeshBuilder` representation or ``None``,
        idiom to check if is a mesh representation (could be empty)::

            if primitive.mesh is not None:
                process(primitive.mesh)

        """
        return None

    @abc.abstractmethod
    def vertices(self) -> Iterable[Vec3]:
        """Yields all vertices of the path/mesh representation as
        :class:`~ezdxf.math.Vec3` objects.

        """
        pass


class EmptyPrimitive(Primitive):
    @property
    def is_empty(self) -> bool:
        return True

    def vertices(self) -> Iterable[Vec3]:
        return []


class ConvertedPrimitive(Primitive):
    """Base class for all DXF entities which store the path/mesh representation
    at instantiation.

    """

    def __init__(self, entity: DXFEntity):
        super().__init__(entity)
        self._convert_entity()

    @abc.abstractmethod
    def _convert_entity(self):
        """This method creates the path/mesh representation."""
        pass

    @property
    def path(self) -> Optional[Path]:
        return self._path

    @property
    def mesh(self) -> Optional[MeshBuilder]:
        return self._mesh

    def vertices(self) -> Iterable[Vec3]:
        if self.path:
            yield from self._path.flattening(self.max_flattening_distance)  # type: ignore
        elif self.mesh:
            yield from self._mesh.vertices  # type: ignore


class CurvePrimitive(Primitive):
    @property
    def path(self) -> Optional[Path]:
        """Create path representation on demand."""
        if self._path is None:
            self._path = make_path(self.entity)
        return self._path

    def vertices(self) -> Iterable[Vec3]:
        # Not faster but more precise, because cubic bezier curves do not
        # perfectly represent elliptic arcs (CIRCLE, ARC, ELLIPSE).
        # SPLINE: cubic bezier curves do not perfectly represent splines with
        # degree != 3.
        yield from self.entity.flattening(self.max_flattening_distance)  # type: ignore


class LinePrimitive(Primitive):
    # TODO: apply thickness if not 0
    @property
    def path(self) -> Optional[Path]:
        """Create path representation on demand."""
        if self._path is None:
            self._path = make_path(self.entity)
        return self._path

    def vertices(self) -> Iterable[Vec3]:
        e = self.entity
        yield e.dxf.start
        yield e.dxf.end


class LwPolylinePrimitive(ConvertedPrimitive):
    # TODO: apply thickness if not 0
    def _convert_entity(self):
        e: "LWPolyline" = cast("LWPolyline", self.entity)
        if e.has_width:  # use a mesh representation:
            # TraceBuilder operates in OCS!
            ocs = e.ocs()
            elevation = e.dxf.elevation
            tb = TraceBuilder.from_polyline(e)
            mb = MeshVertexMerger()  # merges coincident vertices
            for face in tb.faces_wcs(ocs, elevation):
                mb.add_face(face)
            self._mesh = MeshBuilder.from_builder(mb)
        else:  # use a path representation to support bulges!
            self._path = make_path(e)


class PointPrimitive(Primitive):
    @property
    def path(self) -> Optional[Path]:
        """Create path representation on demand.

        :class:`Path` can not represent a point, a :class:`Path` with only a
        start point yields not vertices!

        """
        if self._path is None:
            self._path = Path(self.entity.dxf.location)
        return self._path

    def vertices(self) -> Iterable[Vec3]:
        yield self.entity.dxf.location


class MeshPrimitive(ConvertedPrimitive):
    def _convert_entity(self):
        self._mesh = MeshBuilder.from_mesh(self.entity)


class QuadrilateralPrimitive(ConvertedPrimitive):
    # TODO: apply thickness if not 0
    def _convert_entity(self):
        self._path = make_path(self.entity)


class PolylinePrimitive(ConvertedPrimitive):
    # TODO: apply thickness if not 0
    def _convert_entity(self):
        e: "Polyline" = cast("Polyline", self.entity)
        if e.is_2d_polyline and e.has_width:
            # TraceBuilder operates in OCS!
            ocs = e.ocs()
            elevation = e.dxf.elevation.z
            tb = TraceBuilder.from_polyline(e)
            mb = MeshVertexMerger()  # merges coincident vertices
            for face in tb.faces_wcs(ocs, elevation):
                mb.add_face(face)
            self._mesh = MeshBuilder.from_builder(mb)
        elif e.is_2d_polyline or e.is_3d_polyline:
            self._path = make_path(e)
        else:
            m = MeshVertexMerger.from_polyface(e)
            self._mesh = MeshBuilder.from_builder(m)


class HatchPrimitive(ConvertedPrimitive):
    def _convert_entity(self):
        self._path = make_path(self.entity)


DESCENDER_FACTOR = 0.333  # from TXT SHX font - just guessing
X_HEIGHT_FACTOR = 0.666  # from TXT SHX font - just guessing


def get_font_name(entity: "DXFEntity"):
    font_name = "txt"
    if entity.doc:
        style_name = entity.dxf.style
        style = entity.doc.styles.get(style_name)
        if style:
            font_name = style.dxf.font
    return font_name


class TextLinePrimitive(ConvertedPrimitive):
    def _convert_entity(self):
        """Calculates the rough border path for a single line text.

        Calculation is based on a mono-spaced font and therefore the border
        path is just an educated guess.

        Vertical text generation and oblique angle is ignored.

        """

        def text_rotation():
            if fit_or_aligned and not p1.isclose(p2):
                return (p2 - p1).angle
            else:
                return math.radians(text.dxf.rotation)

        def location():
            if alignment == "LEFT":
                return p1
            elif fit_or_aligned:
                return p1.lerp(p2, factor=0.5)
            else:
                return p2

        text = cast("Text", self.entity)
        if text.dxftype() == "ATTDEF":
            # ATTDEF outside of a BLOCK renders the tag rather than the value
            content = text.dxf.tag
        else:
            content = text.dxf.text

        content = plain_text(content)
        if len(content) == 0:
            # empty path - does not render any vertices!
            self._path = Path()
            return

        p1: Vec3 = text.dxf.insert
        p2: Vec3 = text.dxf.align_point
        font = fonts.make_font(
            get_font_name(text), text.dxf.height, text.dxf.width
        )
        text_line = TextLine(content, font)
        alignment: str = text.get_align()
        fit_or_aligned = alignment == "FIT" or alignment == "ALIGNED"
        if text.dxf.halign > 2:  # ALIGNED=3, MIDDLE=4, FIT=5
            text_line.stretch(alignment, p1, p2)
        halign, valign = unified_alignment(text)
        mirror_x = -1 if text.is_backward else 1
        mirror_y = -1 if text.is_upside_down else 1
        oblique: float = math.radians(text.dxf.oblique)
        corner_vertices = text_line.corner_vertices(
            location(),
            halign,
            valign,
            angle=text_rotation(),
            scale=(mirror_x, mirror_y),
            oblique=oblique,
        )

        ocs = text.ocs()
        self._path = from_vertices(
            ocs.points_to_wcs(corner_vertices),
            close=True,
        )


class MTextPrimitive(ConvertedPrimitive):
    def _convert_entity(self):
        """Calculates the rough border path for a MTEXT entity.

        Calculation is based on a mono-spaced font and therefore the border
        path is just an educated guess.

        Most special features of MTEXT is not supported.

        """

        def get_content() -> List[str]:
            text = mtext.plain_text(split=False)
            return text_wrap(text, box_width, font.text_width)  # type: ignore

        def get_max_str() -> str:
            return max(content, key=lambda s: len(s))

        def get_rect_width() -> float:
            if box_width:
                return box_width
            s = get_max_str()
            if len(s) == 0:
                s = " "
            return font.text_width(s)

        def get_rect_height() -> float:
            line_height = font.measurements.total_height
            cap_height = font.measurements.cap_height
            # Line spacing factor: Percentage of default (3-on-5) line
            # spacing to be applied.

            # thx to mbway: multiple of cap_height between the baseline of the
            # previous line and the baseline of the next line
            # 3-on-5 line spacing = 5/3 = 1.67
            line_spacing = cap_height * mtext.dxf.line_spacing_factor * 1.67
            spacing = line_spacing - line_height
            line_count = len(content)
            return line_height * line_count + spacing * (line_count - 1)

        def get_ucs() -> UCS:
            """Create local coordinate system:
            origin = insertion point
            z-axis = extrusion vector
            x-axis = text_direction or text rotation, text rotation requires
                extrusion vector == (0, 0, 1) or treatment like an OCS?

            """
            origin = mtext.dxf.insert
            z_axis = mtext.dxf.extrusion  # default is Z_AXIS
            x_axis = X_AXIS
            if mtext.dxf.hasattr("text_direction"):
                x_axis = mtext.dxf.text_direction
            elif mtext.dxf.hasattr("rotation"):
                # TODO: what if extrusion vector is not (0, 0, 1)
                x_axis = Vec3.from_deg_angle(mtext.dxf.rotation)
                z_axis = Z_AXIS
            return UCS(origin=origin, ux=x_axis, uz=z_axis)

        def get_shift_factors():
            halign, valign = unified_alignment(mtext)
            shift_x = 0
            shift_y = 0
            if halign == const.CENTER:
                shift_x = -0.5
            elif halign == const.RIGHT:
                shift_x = -1.0
            if valign == const.MIDDLE:
                shift_y = 0.5
            elif valign == const.BOTTOM:
                shift_y = 1.0
            return shift_x, shift_y

        def get_corner_vertices() -> Iterable[Vec3]:
            """Create corner vertices in the local working plan, where
            the insertion point is the origin.
            """
            if columns:
                rect_width = columns.total_width
                rect_height = columns.total_height
                # TODO: this works only for reliable sources like AutoCAD,
                #  BricsCAD and ezdxf! So far no known column support from
                #  other DXF exporters.
            else:
                rect_width = mtext.dxf.get("rect_width", get_rect_width())
                rect_height = mtext.dxf.get("rect_height", get_rect_height())
            # TOP LEFT alignment:
            vertices = [
                Vec3(0, 0),
                Vec3(rect_width, 0),
                Vec3(rect_width, -rect_height),
                Vec3(0, -rect_height),
            ]
            sx, sy = get_shift_factors()
            shift = Vec3(sx * rect_width, sy * rect_height)
            return (v + shift for v in vertices)

        mtext: "MText" = cast("MText", self.entity)
        columns = mtext.columns
        if columns is None:
            box_width = mtext.dxf.get("width", 0)
            font = fonts.make_font(
                get_font_name(mtext), mtext.dxf.char_height, 1.0
            )
            content: List[str] = get_content()
            if len(content) == 0:
                # empty path - does not render any vertices!
                self._path = Path()
                return
        ucs = get_ucs()
        corner_vertices = get_corner_vertices()
        self._path = from_vertices(
            ucs.points_to_wcs(corner_vertices),
            close=True,
        )


class PathPrimitive(Primitive):
    def __init__(
        self, path: Path, entity: DXFEntity, max_flattening_distance=None
    ):
        super().__init__(entity, max_flattening_distance)
        self._path = path

    @property
    def path(self) -> Optional[Path]:
        return self._path

    def vertices(self) -> Iterable[Vec3]:
        yield from self._path.flattening(self.max_flattening_distance)  # type: ignore


class ImagePrimitive(ConvertedPrimitive):
    def _convert_entity(self):
        self._path = make_path(self.entity)


class ViewportPrimitive(ConvertedPrimitive):
    def _convert_entity(self):
        vp = self.entity
        if vp.dxf.status == 0:  # Viewport is off
            return  # empty primitive
        self._path = make_path(vp)


# SHAPE is not supported, could not create any SHAPE entities in BricsCAD
_PRIMITIVE_CLASSES = {
    "3DFACE": QuadrilateralPrimitive,
    "ARC": CurvePrimitive,
    # TODO: ATTRIB and ATTDEF could contain embedded MTEXT,
    #  but this is not supported yet!
    "ATTRIB": TextLinePrimitive,
    "ATTDEF": TextLinePrimitive,
    "CIRCLE": CurvePrimitive,
    "ELLIPSE": CurvePrimitive,
    "HATCH": HatchPrimitive,  # multi-path object
    "MPOLYGON": HatchPrimitive,  # multi-path object
    "HELIX": CurvePrimitive,
    "IMAGE": ImagePrimitive,
    "LINE": LinePrimitive,
    "LWPOLYLINE": LwPolylinePrimitive,
    "MESH": MeshPrimitive,
    "MTEXT": MTextPrimitive,
    "POINT": PointPrimitive,
    "POLYLINE": PolylinePrimitive,
    "SPLINE": CurvePrimitive,
    "SOLID": QuadrilateralPrimitive,
    "TEXT": TextLinePrimitive,
    "TRACE": QuadrilateralPrimitive,
    "VIEWPORT": ViewportPrimitive,
    "WIPEOUT": ImagePrimitive,
}


def make_primitive(
    entity: DXFEntity, max_flattening_distance=None
) -> Primitive:
    """Factory to create path/mesh primitives. The `max_flattening_distance`
    defines the max distance between the approximation line and the original
    curve. Use `max_flattening_distance` to override the default value.

    Returns an **empty primitive** for unsupported entities. The `empty` state
    of a primitive can be checked by the property :attr:`is_empty`.
    The :attr:`path` and the :attr:`mesh` attributes of an empty primitive
    are ``None`` and the :meth:`vertices` method  yields no vertices.

    .. versionchanged:: 0.17
        regular support for the :class:`~ezdxf.entities.Hatch` entity.

    """
    cls = _PRIMITIVE_CLASSES.get(entity.dxftype(), EmptyPrimitive)
    primitive = cls(entity)
    if max_flattening_distance:
        primitive.max_flattening_distance = max_flattening_distance
    return primitive


def recursive_decompose(entities: Iterable[DXFEntity]) -> Iterable[DXFEntity]:
    """Recursive decomposition of the given DXF entity collection into a flat
    DXF entity stream. All block references (INSERT) and entities which provide
    a :meth:`virtual_entities` method will be disassembled into simple DXF
    sub-entities, therefore the returned entity stream does not contain any
    INSERT entity.

    Point entities will **not** be disassembled into DXF sub-entities,
    as defined by the current point style $PDMODE.

    These entity types include sub-entities and will be decomposed into
    simple DXF entities:

        - INSERT
        - DIMENSION
        - LEADER
        - MLEADER
        - MLINE

    Decomposition of XREF, UNDERLAY and ACAD_TABLE entities is not supported.

    """
    for entity in entities:
        if isinstance(entity, Insert):
            if entity.mcount > 1:
                yield from recursive_decompose(entity.multi_insert())
            else:
                yield from entity.attribs
                yield from recursive_decompose(virtual_entities(entity))
        # has a required __virtual_entities__() to be rendered?
        elif isinstance(entity, SupportsVirtualEntities):
            # could contain block references:
            yield from recursive_decompose(virtual_entities(entity))
        else:
            yield entity


def to_primitives(
    entities: Iterable[DXFEntity], max_flattening_distance: float = None
) -> Iterable[Primitive]:
    """Yields all DXF entities as path or mesh primitives. Yields
    unsupported entities as empty primitives, see :func:`make_primitive`.

    Args:
        entities: iterable of DXF entities
        max_flattening_distance: override the default value

    """
    for e in entities:
        yield make_primitive(e, max_flattening_distance)


def to_vertices(primitives: Iterable[Primitive]) -> Iterable[Vec3]:
    """Yields all vertices from the given `primitives`. Paths will be flattened
    to create the associated vertices. See also :func:`to_control_vertices` to
    collect only the control vertices from the paths without flattening.

    """
    for p in primitives:
        yield from p.vertices()


def to_paths(primitives: Iterable[Primitive]) -> Iterable[Path]:
    """Yields all :class:`~ezdxf.path.Path` objects from the given
    `primitives`. Ignores primitives without a defined path.

    """
    for prim in primitives:
        if prim.path is not None:  # lazy evaluation!
            yield prim.path


def to_meshes(primitives: Iterable[Primitive]) -> Iterable[MeshBuilder]:
    """Yields all :class:`~ezdxf.render.MeshBuilder` objects from the given
    `primitives`. Ignores primitives without a defined mesh.

    """
    for prim in primitives:
        if prim.mesh is not None:
            yield prim.mesh


def to_control_vertices(primitives: Iterable[Primitive]) -> Iterable[Vec3]:
    """Yields all path control vertices and all mesh vertices from the given
    `primitives`. Like :func:`to_vertices`, but without flattening.

    """
    for prim in primitives:
        # POINT has only a start point and yields from vertices()!
        if prim.path:
            yield from prim.path.control_vertices()
        else:
            yield from prim.vertices()
